from typing import List, Tuple

import numpy as np
from napari.layers import Layer, Points

from napari_allencell_annotator.view.i_viewer import IViewer


class FakeViewer(IViewer):
    def __init__(self):
        super().__init__()

        self._layers = []
        self.alerts = []

    def add_image(self, image: np.ndarray) -> None:
        self._layers.append(image)

    def clear_layers(self) -> None:
        self._layers.clear()

    def alert(self, alert_msg: str) -> None:
        self.alerts.append(alert_msg)

    def get_layers(self) -> List[Layer]:
        return self._layers

    def get_all_points_layers(self) -> List[Points]:
        return [layer for layer in self.get_layers() if isinstance(layer, Points)]

    def create_points_layer(self, name: str, color: str, visible: bool, data: np.ndarray = None) -> Points:
        points: Points = Points(data=data, name=name, color=color, visible=visible, ndim=6)
        self._layers.append(points)
        return points

    def get_selected_points(self, point_layer: Points) -> List[Tuple]:
        return list(map(tuple, point_layer.data))

    def get_all_point_annotations(self) -> dict[str, list[tuple]]:
        all_point_annotations: dict[str, list[tuple]] = {}

        all_points_layers: list[Points] = self.get_all_points_layers()
        for points_layer in all_points_layers:
            all_point_annotations[points_layer.name] = self.get_selected_points(points_layer)

        return all_point_annotations



from napari_allencell_annotator.model.key import Key


class ComboKey(Key):
    def __init__(self, type: type, dropdown_options: list[str], key_default_value: str = None):
        super().__init__("list", key_default_value)
        self._dropdown_options = dropdown_options

    def get_options(self) -> list[str]:
        return self._dropdown_options



from pathlib import Path
from typing import Optional, Any

from PyQt5.QtCore import QObject
from qtpy.QtCore import Signal

from napari_allencell_annotator.model.key import Key


class AnnotatorModel(QObject):

    image_changed: Signal = Signal()
    image_count_changed: Signal = Signal(int)
    images_shuffled: Signal = Signal(bool)
    image_set_added: Signal = Signal()

    def __init__(self):
        super().__init__()
        # dict of annotation key names -> Key objects containing information about that key
        # such as default values, options, type
        self._annotation_keys: dict[str, Key] = {}
        # Images that have been added to annotator
        self._added_images: list[Path] = []
        # Shuffled images list. If user has not selected shuffle, this remains None
        # and is populated with a shuffled list if the user has selected shuffle
        self._shuffled_images: Optional[list[Path]] = None

        # THE FOLLOWING FIELDS ONLY ARE NEEDED WHEN ANNOTATING STARTS AND ARE INITIALIZED AFTER STARTING.
        # Current image index, which is none by default
        # Changes to curr_img_index through set_curr_img_index() emits an image_changed event which parts of the app
        # react to display that image. -1 if the user has not started annotating.
        self._curr_img_index: int = -1
        self._previous_img_index: int = -1  # index of previously viewed image, -1 by default
        # annotations that have been crated. If annotating has not started, is None by default.
        # dict of annotated image path -> list of annotations for that image
        self._created_annotations: Optional[dict[Path, list[Any]]] = None
        # path to csv where data should be saved.
        # None if annotating has not started.
        self._csv_save_path: Optional[Path] = None

        self._annotation_started = False

    def get_annotation_keys(self) -> dict[str, Key]:
        return self._annotation_keys

    def clear_annotation_keys(self) -> None:
        self._annotation_keys.clear()

    def set_annotation_keys(self, annotation_keys: dict[str, Key]) -> None:
        self._annotation_keys = annotation_keys

    def add_image(self, file_item: Path, idx: Optional[int] = None) -> None:
        if idx:
            self._added_images.insert(idx)
        else:
            self._added_images.append(file_item)
        self.image_count_changed.emit(self.get_num_images())

    def get_all_images(self) -> list[Path]:
        if self.is_images_shuffled():
            return self._shuffled_images
        else:
            return self._added_images

    def get_image_at(self, idx: int) -> Path:
        if self.is_images_shuffled():
            return self._shuffled_images[idx]
        else:
            return self._added_images[idx]

    def get_num_images(self) -> int:
        if self.is_images_shuffled():
            return len(self._shuffled_images)
        else:
            return len(self._added_images)

    def set_all_images(self, list_of_img: list[Path]) -> None:
        self._added_images = list_of_img
        self.image_set_added.emit()
        self.image_count_changed.emit(self.get_num_images())

    def clear_all_images(self) -> None:
        self._added_images = []
        self._created_annotations = None
        if self.is_images_shuffled():
            self.set_shuffled_images(None)
        self.image_count_changed.emit(0)

    def empty_image_list(self) -> None:
        if self.is_images_shuffled():
            self._shuffled_images = []
        else:
            self._added_images = []

    def remove_image(self, item: Path) -> None:
        self._added_images.remove(item)
        self.image_count_changed.emit(self.get_num_images())

    def set_shuffled_images(self, shuffled: Optional[list[Path]]) -> None:
        self._shuffled_images = shuffled
        if shuffled is not None:
            # we are setting the _shuffled_images field to a list of shuffled images. emit event so ui reacts
            self.images_shuffled.emit(True)
        else:
            self.images_shuffled.emit(False)

    def get_shuffled_images(self) -> list[Path]:
        return self._shuffled_images

    def is_images_shuffled(self) -> bool:
        return self._shuffled_images is not None

    def get_curr_img_index(self) -> int:
        return self._curr_img_index

    def set_curr_img_index(self, idx: int) -> None:
        self._curr_img_index = idx
        self.image_changed.emit()

    def get_curr_img(self) -> Optional[Path]:
        if self.get_curr_img_index() != -1:
            if self.is_images_shuffled():
                return self._shuffled_images[self._curr_img_index]
            else:
                return self._added_images[self._curr_img_index]
        else:
            return None

    def get_annotations(self) -> dict[Path, list[Any]]:
        return self._created_annotations

    def add_annotation(self, file_path: Path, annotation: list[Any]):
        self._created_annotations[file_path] = annotation

    def set_annotations(self, annotations: dict[Path, list[Any]]) -> None:
        self._created_annotations = annotations

    def set_previous_image_index(self, idx: int) -> None:
        self._previous_img_index = idx

    def get_previous_image_index(self) -> int:
        return self._previous_img_index

    def set_csv_save_path(self, path: Path) -> None:
        self._csv_save_path = path

    def get_csv_save_path(self) -> Path:
        return self._csv_save_path

    def is_annotation_started(self) -> bool:
        return self._annotation_started

    def set_annotation_started(self, started: bool) -> None:
        self._annotation_started = started



from typing import Any


class Key:
    def __init__(self, type: str, key_default_value: Any = None) -> None:
        self._type: str = type
        self._key_default_value = key_default_value

    def get_type(self) -> str:
        return self._type

    def get_default_value(self) -> Any:
        return self._key_default_value



from qtpy.QtWidgets import QMessageBox


class Popup:
    @classmethod
    def make_popup(cls, text: str) -> bool:
        """
        Pop up dialog to ask the user yes or no.

        Parameters
        ----------
        text : str
            question for the message box.

        Returns
        ----------
        bool
            user input, true if 'Yes' false if 'No'

        """
        msg_box = QMessageBox()
        msg_box.setText(text)
        msg_box.setStandardButtons(QMessageBox.No | QMessageBox.Yes)
        return_value = msg_box.exec()
        if return_value == QMessageBox.Yes:
            return True
        else:
            return False



import pytest

from napari_allencell_annotator._style import Style


class TestStyle:
    @pytest.mark.parametrize("name", [None, "not_a_qt_stylesheet.txt"])
    def test_load_stylesheet_bad_name(self, name):
        with pytest.raises(ValueError):
            Style.get_stylesheet(name)

    def test_load_stylesheet_bad_path(self):
        with pytest.raises(IOError):
            Style.get_stylesheet("not_found.qss")

    def test_load_stylesheet(self):
        contents = Style.get_stylesheet("main.qss")
        assert contents is not None
        assert len(contents) > 0



from typing import List, Tuple, Dict, Optional
from enum import Enum

import dask.array
import numpy as np
from napari.layers import Layer, Points
from napari_allencell_annotator.view.i_viewer import IViewer
from napari.utils.notifications import show_info
import napari


class PointsLayerMode(Enum):
    """
    Mode for view.

    ADD is used to add points.
    SELECT is used to move, edit, or delete points.
    PAN_ZOOM is the default mode and allows normal interactivity with the canvas.
    """

    ADD = "add"
    SELECT = "select"
    PAN_ZOOM = "pan_zoom"


class Viewer(IViewer):
    """Handles actions related to napari viewer"""

    def __init__(self, viewer: napari.Viewer):
        super().__init__()
        self.viewer: napari.Viewer = viewer

    def add_image(self, image: np.ndarray) -> None:
        """
        Add an image to the napari viewer

        Parameters
        ----------
        image: np.ndarray
            An image to be added
        """
        self.viewer.add_image(image)

    def clear_layers(self) -> None:
        """
        Clear all images from the napari viewer
        """
        self.viewer.layers.clear()

    def alert(self, alert_msg: str) -> None:
        """
        Displays an error alert on the viewer.

        Parameters
        ----------
        alert_msg : str
            The message to be displayed
        """
        show_info(alert_msg)

    def get_layers(self) -> List[Layer]:
        """
        Returns a list of all layers in the viewer.
        """
        return list(self.viewer.layers)

    def get_all_points_layers(self) -> List[Points]:
        """
        Returns a list of all point layers in the viewer.
        """
        return [layer for layer in self.get_layers() if isinstance(layer, Points)]

    def create_points_layer(self, name: str, color: str, visible: bool, data: np.ndarray = None) -> Points:
        """
        Creates a new point layer and sets to ADD mode to allow users to select points.

        Parameters
        ----------
        name: str
            The name of the point layer
        color: str
            The face color of the points
        visible: bool
            Whether the point layer is visible in the viewer
        data: np.ndarray = None
            A numpy array of point coordinates

        Returns
        -------
        Points
            A new point layer
        """
        points_layer: Points = self.viewer.add_points(
            data=data, name=name, face_color=color, visible=visible, ndim=self.viewer.dims.ndim
        )
        return points_layer

    def set_points_layer_mode(self, points_layer: Points, mode: PointsLayerMode) -> None:
        """
        Sets a point layer's mode.

        Parameters
        ----------
        points_layer: Points
            The Points layer
        mode: str
            The mode
        """
        points_layer.mode = mode.value

    def get_selected_points(self, point_layer: Points) -> list[tuple]:
        """
        Returns a list of points in the point layer.

        Parameters
        ----------
        point_layer: Points
            The point layer

        Returns
        -------
        List[Tuple[float]]
            A list of tuples representing points in the point layer
        """
        # return ex. [(0, 0, 0, 0, 0, 0), (1, 1, 1, 1, 1, 1)]
        selected_points: List[tuple] = list(map(tuple, point_layer.data))
        return selected_points

    def get_all_point_annotations(self) -> dict[str, list[tuple]]:
        """
        Returns a dictionary of point layer names mapping to a list of selected coordinates.
        """
        all_point_annotations: dict[str, list[tuple]] = {}

        all_points_layers: list[Points] = self.get_all_points_layers()
        for points_layer in all_points_layers:
            all_point_annotations[points_layer.name] = self.get_selected_points(points_layer)

        return all_point_annotations



from typing import Set
from PyQt5.QtWidgets import QDialog

from napari_allencell_annotator.widgets.file_item import FileItem
from napari_allencell_annotator.widgets.scrollable_popup import ScrollablePopup


class FileScrollablePopup:
    @classmethod
    def make_popup(cls, msg: str, checked_files: Set[FileItem]) -> bool:
        """
        Pop up dialog showing currently checked image files to ask the user yes or no.

        Parameters
        ----------
        msg: str
            Question for the message box.
        checked_files: Set[FileItem]
            The list of currently checked files.

        Returns
        ----------
        bool
            user input, true if 'OK' false if 'Cancel'
        """
        names: Set[str] = set()
        for item in checked_files:
            names.add(item.file_path.name)
        msg_box: ScrollablePopup = ScrollablePopup(msg, names)
        return_value: int = msg_box.exec()
        if return_value == QDialog.Accepted:
            return True
        else:
            return False



from unittest import mock
from unittest.mock import MagicMock

from napari_allencell_annotator.widgets.popup import Popup, QMessageBox


class TestPopup:
    def test_popup_yes(self):
        with mock.patch.object(QMessageBox, "__init__", lambda x: None):
            QMessageBox.setText = MagicMock()
            QMessageBox.setStandardButtons = MagicMock()
            QMessageBox.exec = MagicMock(return_value=QMessageBox.Yes)

            assert Popup.make_popup("text")
            QMessageBox.setText.assert_called_once_with("text")
            QMessageBox.setStandardButtons.assert_called_once_with(QMessageBox.No | QMessageBox.Yes)
            QMessageBox.exec.assert_called_once_with()

    def test_popup_no(self):
        with mock.patch.object(QMessageBox, "__init__", lambda x: None):
            QMessageBox.setText = MagicMock()
            QMessageBox.setStandardButtons = MagicMock()
            QMessageBox.exec = MagicMock(return_value=QMessageBox.No)

            assert not Popup.make_popup("text")
            QMessageBox.setText.assert_called_once_with("text")
            QMessageBox.setStandardButtons.assert_called_once_with(QMessageBox.No | QMessageBox.Yes)
            QMessageBox.exec.assert_called_once_with()



import json
from pathlib import Path

from napari_allencell_annotator.model.combo_key import ComboKey
from napari_allencell_annotator.model.key import Key


class JSONUtils:

    @staticmethod
    def dict_to_json_dump(annotation_keys: dict[str, Key]) -> str:
        # convert annotation_keys to json compatible format
        keys_dict: dict[str, dict] = {}
        for key, key_info in annotation_keys.items():
            keys_dict[key] = {
                "type": str(key_info.get_type()),
                "default": key_info.get_default_value(),
            }
            if isinstance(key_info, ComboKey):
                keys_dict[key]["options"] = key_info.get_options()
        return json.dumps(keys_dict)

    @staticmethod
    def json_dump_to_dict(json_data: str) -> dict[str, Key]:
        original_dict = json.loads(json_data)
        converted_dict: dict[str, Key] = {}
        for key, value in original_dict.items():
            # this means it is a combo_key
            if "options" in list(value.keys()):
                converted_dict[key] = ComboKey(value["type"], value["options"], value["default"])
            # is a Key
            else:
                converted_dict[key] = Key(value["type"], value["default"])

        return converted_dict

    @staticmethod
    def get_json_data(path: Path) -> str:
        with open(path, "r") as f:
            return f.read()

    @staticmethod
    def write_json_data(data: str, path: Path) -> None:
        with open(path, "w") as f:
            f.write(data)



from abc import ABC, abstractmethod
import numpy as np
from napari.layers import Layer, Points
from typing import List, Tuple


class IViewer(ABC):
    def __init__(self):
        """Base abstract class for the viewer"""
        super().__init__()

    @abstractmethod
    def add_image(self, image: np.ndarray) -> None:
        pass

    @abstractmethod
    def clear_layers(self) -> None:
        pass

    @abstractmethod
    def alert(self, alert_msg: str) -> None:
        pass

    @abstractmethod
    def get_layers(self) -> List[Layer]:
        pass

    @abstractmethod
    def get_all_points_layers(self) -> List[Points]:
        pass

    @abstractmethod
    def create_points_layer(self, name: str, color: str, visible: bool, data: np.ndarray = None) -> Points:
        pass

    @abstractmethod
    def get_selected_points(self, point_layer: Points) -> List[Tuple]:
        pass

    @abstractmethod
    def get_all_point_annotations(self) -> dict[str, list[tuple]]:
        pass



from enum import Enum
from typing import Any

from qtpy.QtWidgets import QLayout
from qtpy.QtWidgets import QListWidgetItem, QListWidget, QWidget, QHBoxLayout, QLabel
from qtpy.QtCore import Qt


class ItemType(Enum):
    """"""

    STRING = "string"
    NUMBER = "number"
    BOOL = "bool"
    LIST = "list"
    POINT = "point"


class TemplateItem(QListWidgetItem):
    """
    A class used to show an annotation template QListWidgetItems.

    Attributes
    ----------
    type : ItemType
    default : Any
    editable_widget : QWidget
    value
    """

    def __init__(self, parent: QListWidget, name: str, type: ItemType, default: Any, editable_widget: QWidget):
        QListWidgetItem.__init__(self, parent)
        self._type: ItemType = type
        self.default: Any = default
        self.value: Any = default
        self.editable_widget: QWidget = editable_widget
        self.widget = QWidget()
        self.parent = parent

        self.layout = QHBoxLayout()
        self.name = QLabel(name)

        if self.type != ItemType.POINT:
            self.layout.addWidget(self.name)
            self.layout.addWidget(self.editable_widget)
            self.editable_widget.setEnabled(True)
        else:
            self.layout.addWidget(self.name)
            self.layout.addWidget(self.editable_widget)
            self.editable_widget.setFixedWidth(200)
            self.editable_widget.setEnabled(False)

        self.layout.setContentsMargins(2, 12, 8, 12)
        self.layout.setSpacing(2)
        self.layout.setSizeConstraint(QLayout.SetMinimumSize)
        self.widget.setLayout(self.layout)
        self.setSizeHint(self.widget.minimumSizeHint())
        parent.setItemWidget(self, self.widget)

    @property
    def type(self) -> ItemType:
        """Annotation type (string, number, bool, list)"""
        return self._type

    def set_default_value(self):
        """Set the value for the annotation to the default."""
        self.set_value(self.default)

    def set_value(self, val: Any):
        """
        Set the value of the annotation.

        Parameters
        ----------
        val : Any
        """
        if self._type == ItemType.STRING:
            self.editable_widget.setText(val)
        elif self._type == ItemType.NUMBER:
            self.editable_widget.setValue(int(val))
        elif self._type == ItemType.BOOL:
            if isinstance(val, str):
                # val is a str from previous annotation
                self.editable_widget.setChecked(eval(val))
            else:
                # val is bool from default vals
                self.editable_widget.setChecked(val)
        elif self._type == ItemType.LIST:
            self.editable_widget.setCurrentText(val)

    def get_value(self) -> Any:
        """
        Return the current annotation value.

        Returns
        -------
        Any
        """
        if self._type == ItemType.STRING:
            return self.editable_widget.text()
        elif self._type == ItemType.NUMBER:
            return self.editable_widget.value()
        elif self._type == ItemType.BOOL:
            return self.editable_widget.isChecked()
        elif self._type == ItemType.LIST:
            return self.editable_widget.currentText()

    def create_evt_listener(self):
        """Create event listener for editable widget edits to set the current item."""
        if self._type == ItemType.STRING:
            self.editable_widget.textEdited.connect(lambda: self.parent.setCurrentItem(self))
        elif self._type == ItemType.NUMBER:
            self.editable_widget.valueChanged.connect(lambda: self.parent.setCurrentItem(self))
        elif self._type == ItemType.BOOL:
            self.editable_widget.stateChanged.connect(lambda: self.parent.setCurrentItem(self))
        elif self._type == ItemType.LIST:
            self.editable_widget.activated.connect(lambda: self.parent.setCurrentItem(self))

    def set_focus(self):
        """Set the annotating focus on the widget of the current item."""
        if self._type == ItemType.STRING:
            self.editable_widget.setFocus()
        elif self._type == ItemType.NUMBER:
            self.editable_widget.lineEdit().setFocus()
        elif self._type == ItemType.LIST:
            self.editable_widget.showPopup()

    def highlight(self):
        """Highlight the editable widget in blue."""
        style = ""
        if self._type == ItemType.STRING:
            style = """QLineEdit{border: 1px solid #759e78}"""
        elif self._type == ItemType.NUMBER:
            style = """QSpinBox{border: 1px solid #759e78}"""
        elif self._type == ItemType.BOOL:
            style = """QCheckBox:indicator{border: 1px solid #759e78}"""
        elif self._type == ItemType.LIST:
            style = """QComboBox{border: 1px solid #759e78}"""
        self.editable_widget.setStyleSheet(style)

    def unhighlight(self):
        """Unhighlight the editable widget."""
        style = ""
        if self._type == ItemType.STRING:
            style = """QLineEdit{}"""
        elif self._type == ItemType.NUMBER:
            style = """QSpinBox{}"""
        elif self._type == ItemType.BOOL:
            style = """QCheckBox:indicator{}"""
        elif self._type == ItemType.LIST:
            style = """QComboBox{}"""
        self.editable_widget.setStyleSheet(style)



from pathlib import Path

from napari_allencell_annotator.model.annotation_model import AnnotatorModel
from napari_allencell_annotator.model.key import Key
from napari_allencell_annotator.util.file_utils import FileUtils
from napari_allencell_annotator.util.json_utils import JSONUtils
from napari_allencell_annotator.view.annotator_view import (
    AnnotatorView,
    AnnotatorViewMode,
)
import napari

from typing import List, Optional
import csv


class AnnotatorController:
    """
    A class used to control the model and view for annotations.

    Inputs
    ----------
    viewer : napari.Viewer
        a napari viewer where the plugin will be used

    Methods
    -------
    set_annot_json_data(dct : dct: Dict[str, Dict[str, Any]])
        Sets annotation data dictionary.

    set_csv_path(path : str)
        Sets csv file name for writing.

    start_viewing()
        Changes view to VIEW mode and render annotations.

    stop_viewing()
        Changes view to ADD mode, resets annotations, and clears annotation json data.

    start_annotating(num_images: int, dct: Dict[str, List[str]])
        Changes annotation view to annotating mode.

    stop_annotating()
        Resets values from annotating and changes mode to VIEW.

    set_curr_img(curr_img_dict : Dict[str, str])
        Sets the current image and adds the image to annotations_dict.

    record_annotations(prev_img: str)
        Adds the outgoing image's annotation values to the files_and_annots.

    read_json(file_path : str)
        Reads a json file into a dictionary and sets annot_json_data.

    read_csv(file_path : str)
        Reads the first line of a csv file into a dictionary and sets annot_json_data.

    write_to_csv()
        Writes header and annotations to the csv file.
    """

    def __init__(self, model: AnnotatorModel, viewer: napari.Viewer):
        self._annotation_model = model

        # open in view mode
        self.view: AnnotatorView = AnnotatorView(model, viewer)

        self.view.show()

        self.view.cancel_btn.clicked.connect(self.stop_viewing)
        # we want to save the annotation for the image that we just switched off of.
        self._annotation_model.image_changed.connect(
            lambda: self.record_annotations(self._annotation_model.get_previous_image_index())
        )

    def write_json(self, file_path: str):
        """
        Write annotation dictionary to a file.

        file_path : str
            file path for json file to write to.
        """

        json_data: str = JSONUtils.dict_to_json_dump(self._annotation_model.get_annotation_keys())
        JSONUtils.write_json_data(json_data, Path(file_path))

    def start_viewing(self, alr_anntd: Optional[bool] = False):
        """Change view to VIEW mode and render annotations."""
        self.view.set_mode(mode=AnnotatorViewMode.VIEW)
        self.view.render_annotations(
            self._annotation_model.get_annotation_keys()
        )  # TODO fix render_annotations to use annoations dict
        # disable edit button if already annotated is True
        self.view.edit_btn.setEnabled(not alr_anntd)

    def stop_viewing(self):
        """Change view to ADD mode, reset annotations, and clear annotation json data."""
        self.view.set_mode(mode=AnnotatorViewMode.ADD)
        self._annotation_model.clear_annotation_keys()

    def start_annotating(self):
        """
        Change annotation view to annotating mode and create files_and_annots with files.

        Parameters
        ----------
        num_images : int
            The total number of images to be annotated.
        dct : Dict[str, List[str]]
            The files to be used. path -> [name, FMS]
        """
        self.view.set_mode(mode=AnnotatorViewMode.ANNOTATE)

        self.view.annot_list.create_evt_listeners()
        # self.view.annot_list.currentItemChanged.connect(self._curr_item_changed)

    def save_annotations(self):
        """Save current annotation data"""
        # save annotations for file we're on
        self.record_annotations(self._annotation_model.get_curr_img_index())
        self.write_csv()

    def stop_annotating(self):
        """Reset values from annotating and change mode to ADD."""
        # TODO: DO WE WANT TO SAVE ALL IMAGES WITHOUT ANNOTATIONS
        self.record_annotations(self._annotation_model.get_curr_img_index())

        # Save rest of annotations, even if empty
        for idx in range(self._annotation_model.get_num_images()):
            if self._annotation_model.get_image_at(idx) not in self._annotation_model.get_annotations():
                self._annotation_model.add_annotation(self._annotation_model.get_image_at(idx), [])

        self.write_csv()
        # reset optional fields in model to None (pre-annottion state)
        self._annotation_model.set_annotation_started(False)
        self._annotation_model.set_csv_save_path(None)
        self.view.set_mode(AnnotatorViewMode.VIEW)

    def _curr_item_changed(self, current, previous):
        """
        Highlight the new current annotation selection and unhighlight the previous.

        Parameters
        ----------
        current : TemplateItem
        previous : TemplateItem
        """
        if current is not None:
            # test
            current.highlight()
            current.set_focus()
        if previous is not None:
            previous.unhighlight()

    def set_curr_img(self):
        """
        Set the current image and add the image to annotations_dict.

        Changes next button if annotating the last image.

        Parameters

        ----------
        curr_img : Dict[str, str]
            The current image {'File Path' : 'path', 'Row' : str(rownum)}
        """
        if self._annotation_model.get_annotations() is not None:
            path: Path = self._annotation_model.get_curr_img()
            # files_and_annots values are lists File Path ->[File Name, FMS, annot1val, annot2val ...]
            # if the file has not been annotated the list is just length 2 [File Name, FMS]
            if (
                path is None
                or path not in list(self._annotation_model.get_annotations().keys())
                or len(self._annotation_model.get_annotations()[path]) == 0
            ):
                # if the image is un-annotated render the default values or no image is selected
                self.view.render_default_values()
            else:
                # if the image has been annotated render the values that were entered
                # dictionary list [2::] is [annot1val, annot2val, ...]
                self.view.render_values(self._annotation_model.get_annotations()[path])

            # convert row to int
            self.view.display_current_progress()
            # if at the end disable next
            if self._annotation_model.get_curr_img_index() == self._annotation_model.get_num_images() - 1:
                self.view.next_btn.setEnabled(False)
            else:
                self.view.next_btn.setEnabled(True)
            # if at the start disable prev
            if self._annotation_model.get_curr_img_index() == 0:
                self.view.prev_btn.setEnabled(False)
            else:
                self.view.prev_btn.setEnabled(True)

    def record_annotations(self, record_idx: int):
        """
        Add the image's annotation values to the annotation dictionary

        Parameters
        ----------
        record_idx : int
            The index of the image we should save annotations for
        """
        if (
            record_idx != -1 and self._annotation_model.is_annotation_started()
        ):  # ignore recording annotations when loading first image or just starting out
            # we're saving annotation for the image we just switched off of.
            self._annotation_model.add_annotation(
                self._annotation_model.get_all_images()[record_idx], self.view.get_curr_annots()
            )

    def read_json(self, file_path: Path):
        # TODO change param to path
        """
        Read a json file into a dictionary and set annot_json_data.

        Parameters
        ----------
        file_path : str
            file path to json file to read from
        """
        json_dict: dict[str, Key] = JSONUtils.json_dump_to_dict(JSONUtils.get_json_data(Path(file_path)))
        self._annotation_model.set_annotation_keys(json_dict)

    def get_annotations_csv(self, annotations: str):
        # TODO change param to path
        """
        Read the first line of a csv file into a dictionary and set annot_json_data.

        Parameters
        ----------
        annotations: str
            a string of annotation dictionary data from the csv
        """
        json_dict: dict[str, Key] = JSONUtils.json_dump_to_dict(annotations)
        self._annotation_model.set_annotation_keys(json_dict)

    def write_csv(self):
        # TODO put into csv utils class
        """write headers and file info"""
        file = open(self._annotation_model.get_csv_save_path(), "w")
        writer = csv.writer(file)
        writer.writerow(["Shuffled:", self._annotation_model.is_images_shuffled()])
        header: List[str] = ["Annotations:", JSONUtils.dict_to_json_dump(self._annotation_model.get_annotation_keys())]
        writer.writerow(header)

        header = ["File Name", "File Path"]
        for name in self.view.annots_order:
            header.append(name)
        writer.writerow(header)
        for path, annotations in self._annotation_model.get_annotations().items():
            writer.writerow([FileUtils.get_file_name(path), str(path)] + annotations)
        file.close()



import random
from pathlib import Path
from napari_allencell_annotator.constants.constants import SUPPORTED_FILE_TYPES


class FileUtils:
    """
    Handles file and directory related functions.
    """

    @staticmethod
    def get_valid_images_sorted(file_list: list[Path]) -> list[Path]:
        """
        Return a sorted list of paths to files that are not hidden and is either a valid image file or a valid zarr image.

        Parameters
        ----------
        file_list: list[Path]
            A list of paths
        """
        valid_files: list[Path] = []
        for file in file_list:
            # is not hidden
            if not file.name.startswith("."):
                # all supported files including raw zarr
                if FileUtils.is_supported(file):
                    valid_files.append(file)
                # if zarr outer folder was selected instead
                elif FileUtils._is_outer_zarr(file):
                    valid_files.append(FileUtils._get_raw_zarr_from_outer_dir(file))

        return sorted(valid_files, key=FileUtils.get_file_name)

    @staticmethod
    def is_supported(file_path: Path) -> bool:
        """
        Check if the provided file name is a supported file.

        This function checks if the file name extension is in
        the supported file types files.

        Parameters
        ----------
        file_path : Path
            Name of the file to check.

        Returns
        -------
        bool
            True if the file is supported.
        """
        extension: str = file_path.suffix
        return extension in SUPPORTED_FILE_TYPES

    @staticmethod
    def shuffle_file_list(files: list[Path]) -> list[Path]:
        """
        Shuffles the file list.

        Parameters
        ----------
        files: list[Path]
            A file list to be shuffled

        Returns
        -------
        list[Path]
            The shuffled file list
        """
        shuffled_list = files.copy()
        random.shuffle(shuffled_list)
        return shuffled_list

    @staticmethod
    def get_file_name(path: Path) -> str:
        """
        Return the parent file name for zarr. Otherwise, return the file name.

        Parameters
        ----------
        path: Path
            The path to an image file
        """
        if path.suffix == ".zarr":
            return path.parent.name
        else:
            return path.name

    @staticmethod
    def _is_outer_zarr(path: Path) -> bool:
        """
        Return whether a given path is the outer directory of a raw zarr.

        Parameters
        ----------
        path: Path
            The path to a directory
        """
        if list(path.glob("*.zarr")):
            return True
        else:
            return False

    @staticmethod
    def _get_raw_zarr_from_outer_dir(path: Path) -> Path:
        """
        Returns the path to the raw zarr in a given directory

        Parameters
        ----------
        path: Path
            The path to an outer zarr directory
        """
        return path.glob("*.zarr").__next__()



from typing import Any, List

from qtpy import QtWidgets
from qtpy.QtWidgets import QLineEdit, QCheckBox, QComboBox, QSpinBox, QPushButton, QSizePolicy, QListWidget

from napari_allencell_annotator.model.combo_key import ComboKey
from napari_allencell_annotator.model.key import Key
from napari_allencell_annotator.widgets.template_item import TemplateItem, ItemType
from napari_allencell_annotator._style import Style


class TemplateList(QListWidget):
    """
    A class used to create a QListWidget for annotation templates.

    Properties
    ----------
    items : List[TemplateItem]

    """

    def __init__(self):
        QListWidget.__init__(self)

        self.setStyleSheet(Style.get_stylesheet("main.qss"))
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        # todo single selection
        self._items: List[TemplateItem] = []
        self.height: int = 0
        self.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)

    @property
    def items(self) -> List[TemplateItem]:
        """
        Item property.

        Returns
        -------
        List[TemplateItem]
            a list of items.
        """
        return self._items

    def next_item(self):
        """Move the current item down one annotation."""
        curr_row = self.currentRow()
        if curr_row < len(self._items) - 1:
            next_row = curr_row + 1
            self.setCurrentRow(next_row)
        else:
            # if at last start over
            self.setCurrentRow(0)

    def prev_item(self):
        """Move the current item up one annotation."""
        curr_row = self.currentRow()
        if curr_row > 0:
            next_row = curr_row - 1
            self.setCurrentRow(next_row)

    def create_evt_listeners(self):
        """Create annotating event listeners for each item."""
        for item in self.items:
            item.create_evt_listener()

    def clear_all(self):
        """
        Clear all data.

        Reset height, items, list.
        """

        self.clear()
        self._items = []

        self.height = 0

    def add_item(self, name: str, key: Key | ComboKey):
        """
        Add annotation template item from dictionary entries.

        Parameters
        ----------
        name : str
            annotation name.
        dct : Dict[str, Any]
            annotation type, default, and options.
        """
        annot_type = str(key.get_type())
        default: Any = key.get_default_value()
        widget = None
        if annot_type == "string":
            annot_type = ItemType.STRING
            widget = QLineEdit(default)
        elif annot_type == "number":
            annot_type = ItemType.NUMBER
            widget = QSpinBox()
            widget.setValue(default)
        elif annot_type == "bool":
            annot_type = ItemType.BOOL
            widget = QCheckBox()
            widget.setChecked(default)
        elif annot_type == "list":
            annot_type = ItemType.LIST
            widget = QComboBox()
            if isinstance(key, ComboKey):
                for opt in key.get_options():
                    widget.addItem(opt)
            widget.setCurrentText(default)
        elif annot_type == "point":
            annot_type = ItemType.POINT
            widget = QPushButton("Select")

        item = TemplateItem(self, name, annot_type, default, widget)

        self._items.append(item)

        self.height = self.height + item.widget.sizeHint().height()
        self.setMaximumHeight(self.height)



from enum import Enum
from typing import List

from qtpy.QtWidgets import QPushButton
from qtpy.QtWidgets import QHBoxLayout, QWidget, QFileDialog
from qtpy.QtCore import Signal
from pathlib import Path

from napari_allencell_annotator.widgets.ome_zarr_directory_or_file_dialog import OmeZarrDirectoryOrFileDialog


class FileInputMode(Enum):
    DIRECTORY: str = "dir"
    FILE: str = "file"
    CSV: str = "csv"
    JSONCSV: str = "jsoncsv"
    JSON: str = "json"


class FileInput(QWidget):
    """
    A file input Widget that includes a file dialog for selecting a file / directory
    and a text box to display the selected file
    inputs:
        mode (FileInputMode): file dialog selection type to File, Directory, Csv, JSON/Csv, or JSON .
        initial_text (str): text to display in the widget before a file has been selected
    """

    files_selected: Signal = Signal(list)  # signal for multiple files
    dir_selected: Signal = Signal(Path)  # signal for dir selection
    file_selected: Signal = Signal(Path)  # TODO remove this once I refactor

    def __init__(
        self,
        parent: QWidget = None,
        mode: FileInputMode = FileInputMode.FILE,
        placeholder_text: str = None,
    ):
        super().__init__(parent)
        self._mode = mode

        self._input_btn = QPushButton(placeholder_text)
        self._input_btn.clicked.connect(self._select)

        layout: QHBoxLayout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._input_btn)
        self.setLayout(layout)

    @property
    def mode(self) -> FileInputMode:
        return self._mode

    def simulate_click(self) -> None:
        """Simulate a click event to open the file dialog."""
        self._input_btn.clicked.emit()

    def toggle(self, enabled: bool) -> None:
        """
        Enable and un-enable user clicking of the add file button.

        Parameters
        ----------
        enabled : bool
        """
        self._input_btn.setEnabled(enabled)

    def _select(self) -> None:
        if self._mode == FileInputMode.FILE:
            self._select_file()
        elif self._mode == FileInputMode.DIRECTORY:
            self._select_dir()
        elif self._mode == FileInputMode.JSON:
            self._select_json()
        elif self._mode == FileInputMode.CSV:
            self._select_csv()
        elif self._mode == FileInputMode.JSONCSV:
            self._select_csv_or_json()

    def _select_file(self) -> None:
        custom_file_dialog: OmeZarrDirectoryOrFileDialog = OmeZarrDirectoryOrFileDialog(self, "Select a file")

        if custom_file_dialog.exec_() == QFileDialog.Accepted:
            files: List[str] = custom_file_dialog.selectedFiles()
            self.files_selected.emit([Path(file) for file in files])

    def _select_json(self) -> None:
        file_path_str, _ = QFileDialog.getSaveFileName(
            self,
            "Select or create a json file",
            filter="JSON Files (*.json)",
            options=QFileDialog.Option.DontUseNativeDialog | QFileDialog.Option.DontUseCustomDirectoryIcons,
        )
        self.file_selected.emit(Path(file_path_str))

    def _select_dir(self) -> None:
        dir_path_str: str = QFileDialog.getExistingDirectory(
            self,
            "Select a directory",
            options=QFileDialog.Option.DontUseNativeDialog | QFileDialog.Option.DontUseCustomDirectoryIcons,
        )

        if dir_path_str != "":
            self.dir_selected.emit(Path(dir_path_str))

    def _select_csv(self) -> None:
        file_path_str: str
        file_path_str, _ = QFileDialog.getSaveFileName(
            self,
            "Select or create a csv file",
            filter="CSV Files (*.csv)",
            options=QFileDialog.Option.DontUseNativeDialog | QFileDialog.Option.DontUseCustomDirectoryIcons,
        )
        self.file_selected.emit(Path(file_path_str))

    def _select_csv_or_json(self) -> None:
        # JSONCSV
        file_path_str: str
        file_path_str, _ = QFileDialog.getOpenFileName(
            self,
            "Select a .csv or .json file with annotations",
            filter="CSV Files (*.csv) ;; JSON (*.json)",
            options=QFileDialog.Option.DontUseNativeDialog | QFileDialog.Option.DontUseCustomDirectoryIcons,
        )
        self.file_selected.emit(Path(file_path_str))



from pathlib import Path


class ImagesModel:
    def __init__(self):
        # Stores annotations, dictionary of Path -> dict of annotations key/value
        self._annotations: dict[Path, dict[str | any]] = dict()

    def add_image(self, file_item: Path) -> None:
        self._added_images.append(file_item)

    def get_all_images(self) -> list[Path]:
        return self._added_images

    def get_num_images(self) -> int:
        return len(self._added_images)

    def set_all_images(self, list_of_img: list[Path]) -> None:
        self._added_images = list_of_img

    def clear_all_images(self) -> None:
        self._added_images = []
        # TODO: fire signal to disable shuffle and delete

    def remove_image(self, item: Path) -> None:
        self._added_images.remove(item)
        # TODO: if theres nothing left, disable shuffle and delete

    #
    # def set_shuffled_images(self, list_of_img: list[Path]) -> None:
    #     self._shuffled_images = list_of_img
    #
    # def get_shuffled_images(self) -> list[Path]:
    #     return self._shuffled_images



from qtpy.QtWidgets import QListWidget, QAbstractItemView
from qtpy.QtCore import Signal

from napari_allencell_annotator.model.combo_key import ComboKey
from napari_allencell_annotator.model.key import Key
from napari_allencell_annotator.widgets.annotation_item import AnnotationItem
from napari_allencell_annotator._style import Style


class AnnotationWidget(QListWidget):
    """
    A class used to create a QListWidget for annotations that are created.

    """

    # signal emitted when annotation check boxes are selected
    annots_selected = Signal(bool)

    def __init__(self):
        QListWidget.__init__(self)
        self.num_checked: int = 0
        # allow drag and drop rearrangement
        self.setDragDropMode(QAbstractItemView.InternalMove)
        self.setStyleSheet(Style.get_stylesheet("main.qss"))

        # TODO: styling https://blog.actorsfit.com/a?ID=01450-929cf741-2d80-418c-8a55-a52395053369

    def clear_all(self):
        """Clear all image data."""
        self.num_checked = 0
        self.clear()

    def add_existing_item(self, name: str, key_info: Key | ComboKey):
        """
        Add a previously created annotation item to the list for editing.

        Parameters
        ----------
        name: str
            a name for the annotation
        dct: Dict[str, Any]
            a dictionary containing type, default, options
        """
        # TODO support float?
        item: AnnotationItem = self.add_new_item()
        annot_type: str = str(key_info.get_type())
        if annot_type == "string":
            item.fill_vals_text(name, key_info.get_default_value())
        elif annot_type == "number":
            item.fill_vals_number(name, key_info.get_default_value())
        elif annot_type == "bool":
            item.fill_vals_check(name, key_info.get_default_value())
        elif annot_type == "list":
            item.fill_vals_list(name, key_info.get_default_value(), key_info.get_options())
        elif annot_type == "point":
            item.fill_vals_point(name)

    def add_new_item(self) -> AnnotationItem:
        """
        Adds a new Annotation Item to the list. .

        Only allows 10 items to be added.

        Returns
        -------
        AnnotationItem
            the item added
        """
        if self.count() < 10:
            item = AnnotationItem(self)
            item.check.stateChanged.connect(lambda: self._check_evt(item))
            h = item.sizeHint().height()
            self.setMaximumHeight(h * self.count())
            return item

    def remove_item(self, item: AnnotationItem):
        """
        Remove the item.

        Parameters
        -------
        item: AnnotationItem
            an item to remove.
        """
        h = item.sizeHint().height()
        self.takeItem(self.row(item))
        self.setMaximumHeight(h * self.count())

    def delete_checked(self):
        """
        Delete the checked items.

        This function emits a annots_selected signal.
        """
        lst = []
        for x in range(self.count()):
            if self.item(x).check.isChecked():
                lst.append(self.item(x))
        for item in lst:
            self.remove_item(item)
        self.num_checked = 0
        self.annots_selected.emit(False)

    def _check_evt(self, item: AnnotationItem):
        """
        Update checked count and emit files_selected signal.

        Params
        -------
        item: AnnotationItem
            the item that has been checked or unchecked.
        """
        if item.check.isChecked():
            self.num_checked = self.num_checked + 1
            if self.num_checked == 1:
                self.annots_selected.emit(True)
        elif not item.check.isChecked():
            self.num_checked = self.num_checked - 1
            if self.num_checked == 0:
                self.annots_selected.emit(False)



from enum import Enum
from typing import Dict, List, Any

from qtpy.QtWidgets import QFrame
from qtpy import QtCore
from qtpy.QtWidgets import (
    QWidget,
    QHBoxLayout,
    QLabel,
    QGridLayout,
    QScrollArea,
    QPushButton,
    QVBoxLayout,
)
from napari import Viewer

from napari_allencell_annotator.model.annotation_model import AnnotatorModel
from napari_allencell_annotator.model.key import Key
from napari_allencell_annotator.widgets.file_input import (
    FileInput,
    FileInputMode,
)
from napari_allencell_annotator.widgets.template_list import TemplateList
from napari_allencell_annotator._style import Style


class AnnotatorViewMode(Enum):
    """
    Mode for view.

    ADD is used when there is not an annotation set selected
    VIEW is used when an annotation set has been made/selected, but annotating has not started.
    ANNOTATE is used when the image set is finalized and annotating has started.
    """

    ADD = "add"
    VIEW = "view"
    ANNOTATE = "annotate"


class AnnotatorView(QFrame):
    """
    A class used to create a view for annotations.

    Inputs
    ----------
    viewer : napari.Viewer
        a napari viewer where the plugin will be used
    mode : AnnotatorViewMode
        a mode for the view

    Methods
    -------
    mode() -> AnnotatorViewMode

    set_mode(mode: AnnotatorViewMode)

    set_num_images(num: int)
        Sets the total images to be annotated.

    set_curr_index(num: int)
        Sets the index of the currently selected image.

    reset_annotations()
        Resets annotation data to empty.

    render_default_values()
        Sets annotation widget values to default.

    render_values(vals: List):
        Sets the values of the annotation widgets to vals.

    get_curr_annots() -> List
        Returns the current annotation values in a list form.

    render_annotations(data : Dict[str,Dict]))
        Renders GUI elements from the dictionary of annotations.
    """

    def __init__(
        self,
        model: AnnotatorModel,
        viewer: Viewer,
        mode: AnnotatorViewMode = AnnotatorViewMode.ADD,
    ):
        super().__init__()
        self._annotator_model = model
        self._annotator_model.image_changed.connect(self._handle_image_changed)
        self._mode = mode
        label = QLabel("Annotations")
        label.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.layout = QVBoxLayout()
        self.layout.addWidget(label)
        self.setStyleSheet(Style.get_stylesheet("main.qss"))
        self.annot_list = TemplateList()
        self.scroll = QScrollArea()
        self.scroll.setWidget(self.annot_list)
        self.scroll.setWidgetResizable(True)
        self.scroll.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.scroll.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        # self.scroll.setSizeAdjustPolicy(QAbstractScrollArea.AdjustToContents)
        self.scroll.setStyleSheet(
            """QScrollBar:vertical {
            width:10px;    
            margin: 0px 0px 0px 0px;
        }"""
        )
        style = """QScrollBar::handle:vertical {border: 0px solid red; border-radius: 
        2px;} """
        self.scroll.setStyleSheet(self.scroll.styleSheet() + style)

        self.layout.addWidget(self.scroll)

        self.curr_index: int = None

        # Add widget visible in ADD mode
        self.add_widget = QWidget()
        add_layout = QHBoxLayout()
        self.create_btn = QPushButton("Create Template")
        self.create_btn.setEnabled(True)
        self.import_btn = QPushButton("Import Template (.csv or .json)")
        self.import_btn.setEnabled(True)
        self.annot_input = FileInput(mode=FileInputMode.JSONCSV, placeholder_text="Start Annotating")
        self.annot_input.toggle(False)

        add_layout.addWidget(self.create_btn, stretch=2)
        add_layout.addWidget(self.import_btn, stretch=2)
        self.add_widget.setLayout(add_layout)
        self.layout.addWidget(self.add_widget)

        # view widget visible in VIEW mode
        self.view_widget = QWidget()
        view_layout = QHBoxLayout()
        self.edit_btn = QPushButton("Edit")
        self.edit_btn.setToolTip("Edit annotation Template")
        self.cancel_btn = QPushButton("Clear")
        self.cancel_btn.setToolTip("Clear annotation template")
        self.cancel_btn.clicked.connect(self._reset_annotations)
        self.start_btn = QPushButton("Start")
        self.start_btn.setToolTip("Start Annotating Images")
        self.csv_input = FileInput(mode=FileInputMode.CSV)
        self.csv_input.toggle(False)
        self.save_json_btn = FileInput(mode=FileInputMode.JSON, placeholder_text="Save")
        self.save_json_btn.toggle(True)
        self.save_json_btn.setToolTip("Save annotation template \n to JSON file")
        self.start_btn.setEnabled(True)
        self.edit_btn.setEnabled(False)
        view_layout.addWidget(self.cancel_btn, stretch=1)
        view_layout.addWidget(self.edit_btn, stretch=1)
        view_layout.addWidget(self.save_json_btn, stretch=1)
        view_layout.addWidget(self.start_btn, stretch=1)
        self.view_widget.setLayout(view_layout)

        # annot widget visible in ANNOTATE mode
        self.annot_widget = QWidget()
        annot_layout = QGridLayout()
        self.save_btn = QPushButton("Save")
        self.exit_btn = QPushButton("Exit")

        self.prev_btn = QPushButton("< Previous")
        self.next_btn = QPushButton("Next >")
        self.next_btn.setEnabled(True)
        self.progress_bar = QLabel()
        self.progress_bar.setAlignment(QtCore.Qt.AlignCenter)
        annot_layout.addWidget(self.progress_bar, 0, 1, 1, 2)
        annot_layout.addWidget(self.save_btn, 1, 1, 1, 1)
        annot_layout.addWidget(self.exit_btn, 1, 0, 1, 1)
        annot_layout.addWidget(self.prev_btn, 1, 2, 1, 1)
        annot_layout.addWidget(self.next_btn, 1, 3, 1, 1)
        self.annot_widget.setLayout(annot_layout)

        self._display_mode()

        self.annots_order: List[str] = []
        self.setLayout(self.layout)
        self.viewer: Viewer = viewer

    @property
    def mode(self) -> AnnotatorViewMode:
        return self._mode

    def set_mode(self, mode: AnnotatorViewMode):
        """
        Set current mode.

        Parameters
        ----------
        mode: AnnotatorViewMode
        """
        self._mode = mode
        self._display_mode()

    def _handle_image_changed(self):
        if self._annotator_model.get_curr_img_index() == -1:
            self.render_default_values()

    def display_current_progress(self):
        """
        display current progres.
        """
        self.progress_bar.setText(
            "{} of {} Images".format(
                self._annotator_model.get_curr_img_index() + 1, self._annotator_model.get_num_images()
            )
        )

    def _reset_annotations(self):
        """Reset annotation data to empty."""
        self.annot_list.clear_all()

        self.annots_order: List[str] = []
        self.scroll.setMaximumHeight(600)

    def render_default_values(self):
        """Set annotation widget values to default."""
        # for curr index if annots exist fill else fill with default
        for item in self.annot_list.items:
            item.set_default_value()

        # TODO why do we need this?
        # self.annot_list.setCurrentItem(self.annot_list.items[0])

    def render_values(self, vals: list[Any]):
        """
        Set the values of the annotation widgets.

        Parameters
        ----------
        vals:List[str]
            the values for the annotations.
        """
        for item, annotation in zip(self.annot_list.items, vals):
            item_data = annotation
            if item_data is None or item_data == "":
                item.set_default_value()
            else:
                item.set_value(item_data)
        # self.annot_list.setCurrentItem(self.annot_list.items[0])

    def get_curr_annots(self) -> List[Any]:
        """
        Return the current annotation values in a list.

        Returns
        ----------
        List
            a list of annotation values.
        """
        annots = []
        for i in self.annot_list.items:
            annots.append(i.get_value())
        return annots

    def _display_mode(self):
        """Render GUI buttons visible depending on the mode."""
        item = self.layout.itemAt(self.layout.count() - 1)
        item.widget().hide()
        self.layout.removeItem(item)
        if self._mode == AnnotatorViewMode.ADD:
            self.add_widget.show()
            self.layout.addWidget(self.add_widget)
        elif self._mode == AnnotatorViewMode.VIEW:
            self.save_json_btn.setEnabled(True)
            self.view_widget.show()
            self.layout.addWidget(self.view_widget)
        elif self._mode == AnnotatorViewMode.ANNOTATE:
            self.annot_widget.show()
            self.layout.addWidget(self.annot_widget)
            self.prev_btn.setEnabled(False)

    def render_annotations(self, data: Dict[str, Key]):
        """
        Read annotation dictionary into individual annotations.

        Parameters
        ----------
        data : Dict[str, Dict[str, Any]]
            The dictionary of annotation names -> a dictionary of types, defaults, and options.
        """
        self.annot_list.clear_all()
        self.annots_order.clear()
        for name, key_info in data.items():
            self._create_annot(name, key_info)
        self.scroll.setMaximumHeight(self.annot_list.height)

    def _create_annot(self, name: str, key: Key):
        """
        Create annotation widgets from dictionary entries.

        Parameters
        ----------
        name : str
            annotation name.
        dct : Dict[str, Any]
            annotation type, default, and options.
        """
        self.annots_order.append(name)
        self.annot_list.add_item(name, key)



import csv
from pathlib import Path

from napari_allencell_annotator.model.annotation_model import AnnotatorModel
from napari_allencell_annotator.util.file_utils import FileUtils
from napari_allencell_annotator.view.images_view import ImagesView
from qtpy import QtCore
from qtpy.QtWidgets import QFrame, QShortcut
from qtpy.QtWidgets import QVBoxLayout, QDialog
from qtpy.QtGui import QKeySequence

from napari_allencell_annotator.controller.annotator_controller import AnnotatorController
from napari_allencell_annotator.widgets.create_dialog import CreateDialog
from napari_allencell_annotator.widgets.template_item import ItemType, TemplateItem
from napari_allencell_annotator.widgets.popup import Popup
from napari_allencell_annotator.view.viewer import Viewer
from napari_allencell_annotator.view.i_viewer import IViewer

import napari
from typing import List, Union


class MainView(QFrame):
    """
    Main plugin view displayed.

    Methods
    -------
    _start_annotating_clicked()
        Verifies that images are added and user wants to proceed, then opens a .csv file dialog.
    stop_annotating()
         Stops annotating in images and annotations views.
    _next_image_clicked()
        Moves to the next image for annotating.
    _prev_image_clicked()
        Moves to the previous image for annotating.
    """

    def __init__(self, napari_viewer: napari.Viewer):
        super().__init__()
        # init viewer and parts of the plugin
        self._viewer: IViewer = Viewer(napari_viewer)
        self._annotator_model = AnnotatorModel()

        # ImagesView and Controller
        self._images_view = ImagesView(self._annotator_model, self._viewer)
        self._images_view.show()

        self.annots = AnnotatorController(self._annotator_model, self._viewer)

        # set layout and add sub views
        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self._images_view, stretch=1)
        self.layout().addWidget(self.annots.view, stretch=1)

        # key shortcuts
        self._shortcut_key_next = QShortcut(QKeySequence(QtCore.Qt.CTRL + QtCore.Qt.Key_Greater), self)
        self._shortcut_key_prev = QShortcut(QKeySequence(QtCore.Qt.CTRL + QtCore.Qt.Key_Less), self)
        self._shortcut_key_down = QShortcut(QKeySequence(QtCore.Qt.CTRL + QtCore.Qt.Key_Return), self)
        self._shortcut_key_up = QShortcut(QKeySequence(QtCore.Qt.CTRL + QtCore.Qt.SHIFT + QtCore.Qt.Key_Return), self)
        self._shortcut_key_check = QShortcut(QKeySequence(QtCore.Qt.CTRL + QtCore.Qt.SHIFT + QtCore.Qt.Key_Space), self)

        self._connect_slots()
        self.show()

    def _connect_slots(self):
        """Connects annotator view buttons to slots"""
        self.annots.view.start_btn.clicked.connect(self._start_annotating_clicked)

        self.annots.view.next_btn.clicked.connect(self._next_image_clicked)
        self.annots.view.prev_btn.clicked.connect(self._prev_image_clicked)
        self.annots.view.csv_input.file_selected.connect(self._csv_write_selected_evt)
        self.annots.view.save_btn.clicked.connect(self._save)
        self.annots.view.exit_btn.clicked.connect(self._exit_clicked)
        self.annots.view.import_btn.clicked.connect(self._import_annots_clicked)
        self.annots.view.annot_input.file_selected.connect(self._csv_json_import_selected_evt)
        self.annots.view.create_btn.clicked.connect(self._create_clicked)
        self.annots.view.save_json_btn.file_selected.connect(self._json_write_selected_evt)
        self.annots.view.edit_btn.clicked.connect(self._create_clicked)

        # handle events
        self._annotator_model.image_changed.connect(self._image_selected)

    def _create_clicked(self):
        """Create dialog window for annotation creation and start viewing on accept event."""

        dlg = CreateDialog(self._annotator_model, parent=self)
        if dlg.exec() == QDialog.Accepted:
            self.annots.start_viewing()

    def _json_write_selected_evt(self, file_path: Path):
        """
        Set json file name and write the annotations to the file.

        Ensure that all file names have .json extension and that a
        file name is selected. Disable save json button.

        Parameters
        ----------
        file_list : List[str]
            The list containing one file name.
        """
        extension = file_path.suffix
        if extension != ".json":
            file_path = file_path.with_suffix(".json")
        self.annots.view.save_json_btn.setEnabled(False)
        self.annots.write_json(file_path)

    def _csv_json_import_selected_evt(self, file_path: Path):
        """
        Read annotations from csv or json file. Read/Save csv images and annotation data if user chooses to.

        Parameters
        ----------
        file_list : List[str]
            The list containing one file name.
        """
        # todo bad file: json or csv --> send back to add
        if file_path is None:
            self._viewer.alert("No selection provided")
        else:
            use_annots: bool = False
            if file_path.suffix == ".json":
                self.annots.read_json(file_path)

            elif file_path.suffix == ".csv":
                use_annots = Popup.make_popup(
                    "Would you like to use the images and annotation values from "
                    "this csv in addition to the annotation template?\n\n "
                    "\n Note: any currently listed images will be cleared."
                )
                file = open(file_path)
                reader = csv.reader(file)
                shuffled: bool = self.str_to_bool(next(reader)[1])
                # annotation data header
                annts = next(reader)[1]
                # set annotation Key dict in model with json header info
                self.annots.get_annotations_csv(annts)
                # skip actual header
                next(reader)
                image_list: list[Path] = []
                if use_annots:
                    # init annotations dict and fill with data from csv
                    self._annotator_model.set_annotations({})

                for row in reader:
                    # for each line, add data to already annotated and check if there are null values
                    # if null, starting row for annotations is set
                    path: Path = Path(row[1])
                    image_list.append(path)
                    if use_annots:
                        self._annotator_model.add_annotation(path, row[2:])
                    # self._images_view.add_new_item(path)
                # start at row 0 if annotation data was not used from csv
                file.close()
                self._annotator_model.set_all_images(image_list)
                if shuffled:
                    self._annotator_model.set_shuffled_images(
                        FileUtils.shuffle_file_list(self._annotator_model.get_all_images())
                    )
                else:
                    self._annotator_model.set_shuffled_images(None)

            self.annots.start_viewing(use_annots)

    def _shuffle_toggled(self, checked: bool):
        """
        Set has_new_shuffled_order to True if images are shuffled.

        Slot function used to record if csv imported image lists
        are given a new shuffle order before annotating. Slot disconnected once
        images have been shuffled once.

        Parameters
        ----------
        checked : bool
            True if shuffle button is pressed "Shuffle and Hide" mode. False if pressed in "Unhide" mode.
        """
        if checked:
            # images have been shuffled, have to adjust order
            self._annotator_model.set_images_shuffled(True)
            self._images_view.shuffle.toggled.disconnect(self._shuffle_toggled)

    def str_to_bool(self, string) -> bool:
        """
        Convert a string to a bool.

        Parameters
        ----------
        string_ : str

        Returns
        -------
        boolean : bool
        """
        if string.lower() == "true":
            return True
        elif string.lower() == "false":
            return False
        else:
            raise ValueError("The value '{}' cannot be mapped to boolean.".format(string))

    def _import_annots_clicked(self):
        """Open file widget for importing csv/json."""
        self.annots.view.annot_input.simulate_click()

    def _csv_write_selected_evt(self, file_path: Path):
        """
        Set csv file name for writing annotations and call _setup_annotating.

        Ensure that all file names have .csv extension and that a
        file name is selected.

        Parameters
        ----------
        file_list : List[str]
            The list containing one file name.
        """
        extension = file_path.suffix
        if extension != ".csv":
            file_path = file_path.with_suffix(".csv")
        self._annotator_model.set_csv_save_path(file_path)
        self._setup_annotating()

    def _start_annotating_clicked(self):
        """
        Verify that images are added and user wants to proceed, then
        open a .csv file dialog.

        Alert user if there are no files added.
        """
        if self._annotator_model.get_num_images() < 1:
            self._viewer.alert("Can't Annotate Without Adding Images")
        else:
            proceed: bool = Popup.make_popup(
                "Once annotating starts both the image set and annotations cannot be "
                "edited.\n Would "
                "you like to continue?"
            )
            if proceed:
                self.annots.view.csv_input.simulate_click()

    def _stop_annotating(self):
        """
        Stop annotating in images and annotations views.

        Display images and annots views.
        """
        self.layout().addWidget(self._images_view, stretch=1)
        self.layout().addWidget(self.annots.view, stretch=1)
        self._images_view.show()
        self.annots.stop_annotating()

        self._images_view.input_file.show()
        self._images_view.input_dir.show()
        self._images_view.shuffle.show()
        self._images_view.delete.show()
        self.annotating_shortcuts_off()

    def _setup_annotating(self):
        """
        Remove images view if shuffled/hidden annotating and start annotation.

        Pass in annotation values if there are any.
        """
        starting_idx: int = 0
        # init annotations dictionary to store data
        if self._annotator_model.get_annotations() is None:
            # there aren't preloaded annotations from a csv, so create an empty set
            self._annotator_model.set_annotations({})

        else:
            # we have preloaded annotations from a csv, reorder so already annotated images are at the front

            old_images_list: list[Path] = self._annotator_model.get_all_images()
            new_images_list: list[Path] = []
            self._annotator_model.empty_image_list()  # reset images list
            for annot_path, annot_list in self._annotator_model.get_annotations().items():
                # if the image with the existing annotations exists
                if annot_path in old_images_list:
                    # if the annotation is complete move to front
                    if annot_list and len(annot_list) == len(self._annotator_model.get_annotation_keys()):
                        new_images_list.insert(starting_idx, annot_path)
                        old_images_list.remove(annot_path)
                        starting_idx = starting_idx + 1
                    else:
                        # otherwise add to back
                        new_images_list.append(annot_path)
                        old_images_list.remove(annot_path)

            # use all images if not shuffled, shuffled image list if it is shuffled
            if not self._annotator_model.is_images_shuffled():
                self._annotator_model.set_all_images(new_images_list + old_images_list)
            else:
                self._annotator_model.set_shuffled_images(new_images_list + old_images_list)

            # if all images annotated, start at 1 minus index
            if starting_idx >= self._annotator_model.get_num_images():
                starting_idx = starting_idx - 1

        self.annotating_shortcuts_on()
        if self._annotator_model.is_images_shuffled():
            # remove file list if blind annotation
            self.layout().removeWidget(self._images_view)
            self._images_view.hide()
            self.annots.start_annotating()

        # TODO: CODE TO READ ANNOTATIONS IF ALREADY EXISTS
        else:
            # start annotating from beginning with just file info
            self._images_view.start_annotating()
            self.annots.start_annotating()

        # remove this
        # self.annots.set_curr_img()
        if not self._annotator_model.is_images_shuffled():
            # alter images view to fit annotation mode
            self._images_view.hide_image_paths()

        # set first image
        self._annotator_model.set_previous_image_index(self._annotator_model.get_curr_img_index())
        self._annotator_model.set_curr_img_index(starting_idx)
        self._annotator_model.set_annotation_started(True)

    def annotating_shortcuts_on(self):
        """Create annotation keyboard shortcuts and connect them to slots."""

        self._shortcut_key_next.activated.connect(self._next_image_clicked)
        self._shortcut_key_prev.activated.connect(self._prev_image_clicked)
        self._shortcut_key_down.activated.connect(self.annots.view.annot_list.next_item)
        self._shortcut_key_up.activated.connect(self.annots.view.annot_list.prev_item)
        self._shortcut_key_check.activated.connect(self._toggle_check)

    def annotating_shortcuts_off(self):
        """Disconnect signals and slots for annotation shortcuts"""
        self._shortcut_key_next.activated.disconnect(self._next_image_clicked)
        self._shortcut_key_prev.activated.disconnect(self._prev_image_clicked)
        self._shortcut_key_down.activated.disconnect(self.annots.view.annot_list.next_item)
        self._shortcut_key_up.activated.disconnect(self.annots.view.annot_list.prev_item)
        self._shortcut_key_check.activated.disconnect(self._toggle_check)

    def _toggle_check(self):
        """Toggle the checkbox state if the current annotation is a checkbox."""
        curr: TemplateItem = self.annots.view.annot_list.currentItem()
        if curr is not None and curr.type == ItemType.BOOL:
            curr.editable_widget.setChecked(not curr.get_value())

    # def _fix_csv_annotations(self, dct: Dict[str, List[str]]):
    #     """
    #     Change csv_annotation_values to reflect any edits to the image list.
    #
    #     Image list could have been shuffled or had items added/deleted.
    #
    #     Parameters
    #     ----------
    #     dct : Dict[str, List[str]]
    #         the image list dictionary file path -> [file name, fms]
    #     """
    #     dct_keys = dct.keys()
    #     alr_anntd_keys = self.csv_annotation_values.keys()
    #     # dct_keys is the dictionary from the images list. may have been edited
    #     # alr_anntd_keys was read in from csv. has not been edited
    #
    #     # remove this
    #     if not list(map(lambda x: str(x), dct_keys)) == alr_anntd_keys:
    #         # if dct .keys is not equal (order not considered) to aa.keys
    #         # means files were either added/deleted
    #         if self.has_new_shuffled_order:
    #             # if dct .keys is not equal (minus order ) to aa.keys. and SHUFFLED
    #             self._unequal_shuffled_fix_csv_annotations(dct)
    #
    #         else:
    #             self._unequal_unshuffled_fix_csv_annotations(dct)
    #     else:
    #         # dct keys and aakeys are equal (except for order)
    #         if self.has_new_shuffled_order:
    #             self._equal_shuffled_fix_csv_annotations(dct)
    #         # if dct.keys == aakeys and NOT SHUFFLED then no changes

    # def _unequal_unshuffled_fix_csv_annotations(self, dct: Dict[str, List[str]]):
    #     """
    #     Change csv_annotation_values to reflect any edits insertions/deletions to the image list.
    #
    #     Find a new starting row if old starting row image was deleted or if new un-annotated images
    #     are the first un-annotated images.
    #
    #     Parameters
    #     ----------
    #     dct : Dict[str, List[str]]
    #         the image list dictionary file path -> [file name, fms]
    #     """
    #     dct_keys = dct.keys()
    #     alr_anntd_keys = self.csv_annotation_values.keys()
    #     new_starting_row_found: bool = False
    #     new_csv_annotations = {}
    #     # order has not been changed/shuffled since upload
    #     # need to add/delete files from already annotated and get a new starting row in case the
    #     # file deleted or added is now the first file with a null annotation
    #     for old_file, row in zip(alr_anntd_keys, range(len(alr_anntd_keys))):
    #         # remove this
    #         if old_file in list(map(lambda x: str(x), dct_keys)):
    #             # old file wasn't removed from files
    #             new_csv_annotations[old_file] = self.csv_annotation_values[old_file]
    #             if not new_starting_row_found:
    #                 if row >= self.starting_row:
    #                     # every file at index before the original starting row was fully annotated
    #                     # if we are past that point in csv_annotation_values then we need to test if we
    #                     # have found a new none annotation value
    #                     if self.has_none_annotation(self.csv_annotation_values[old_file][2::]):
    #                         self.starting_row = len(new_csv_annotations) - 1
    #                         new_starting_row_found = True
    #
    #         # if old_file not in dct_keys dont add it
    #     if len(new_csv_annotations) < len(dct):
    #         # items were added into dct that were not in already annotated
    #         if not new_starting_row_found:
    #             # start on first new item from dct which has not been annotated
    #             self.starting_row = len(new_csv_annotations)
    #             new_starting_row_found = True
    #         for file in list(dct_keys)[len(new_csv_annotations) : :]:
    #             # remove this
    #             new_csv_annotations[file] = dct[file]
    #
    #     if not new_starting_row_found:
    #         self.starting_row = len(new_csv_annotations) - 1
    #     self.csv_annotation_values = new_csv_annotations

    # def _unequal_shuffled_fix_csv_annotations(self, dct: Dict[str, List[str]]):
    #     """
    #     Change csv_annotation_values to reflect any edits insertions/deletions to the image list and shuffling.
    #
    #     Find a new starting row if old starting row image was deleted or if new un-annotated images
    #     are the first un-annotated images.
    #
    #     Reorder the csv_annotation_values keys so that the GUI will read/write
    #     the csv image list in this new shuffled order. This will help with saving progress
    #     when blindly annotating a large image set.
    #
    #     Parameters
    #     ----------
    #     dct : Dict[str, List[str]]
    #         the image list dictionary file path -> [file name, fms]
    #     """
    #     # it has been either shuffled and is now blind or it was given a new shuffle order
    #     # want to save this order in case csv has unshuffled image order, annotation is supposed to be blind
    #     # and next time the csv is opened it will be in insertion order still
    #     dct_keys = dct.keys()
    #     alr_anntd_keys = self.csv_annotation_values.keys()
    #     new_starting_row_found: bool = False
    #     new_csv_annotations = {}
    #     # remove this
    #     for new_file, dct_index in zip(list(map(lambda x: str(x), dct_keys)), range(len(dct_keys))):
    #
    #         if new_file not in alr_anntd_keys:
    #             # only possible to encounter a new file in middle when shuffling has happened
    #             # file added to dct
    #             # remove this
    #             new_csv_annotations[new_file] = dct[Path(new_file)]
    #
    #             if not new_starting_row_found:
    #                 # just added a new, unannotated file
    #                 self.starting_row = dct_index
    #                 new_starting_row_found = True
    #
    #         elif new_file in alr_anntd_keys:
    #             new_csv_annotations[new_file] = self.csv_annotation_values[new_file]
    #             if not new_starting_row_found:
    #                 # todo: could try an optimize this by storing csv_annotation_values indexes
    #                 if self.has_none_annotation(self.csv_annotation_values[new_file][2::]):
    #                     self.starting_row = dct_index
    #                     new_starting_row_found = True
    #     if not new_starting_row_found:
    #         self.starting_row = len(new_csv_annotations) - 1
    #     self.csv_annotation_values = new_csv_annotations
    #
    # def _equal_shuffled_fix_csv_annotations(self, dct: Dict[str, List[str]]):
    #     """
    #     Change csv_annotation_values to reflect shuffling.
    #
    #     Parameters
    #     ----------
    #     dct : Dict[str, List[str]]
    #         the image list dictionary file path -> [file name, fms]
    #     """
    #     dct_keys = dct.keys()
    #     new_starting_row_found: bool = False
    #     new_csv_annotations = {}
    #     # remove this
    #     for new_file, dct_index in itertools.zip_longest(list(map(lambda x: str(x), dct_keys)), range(len(dct_keys))):
    #         new_csv_annotations[new_file] = self.csv_annotation_values[new_file]
    #         if not new_starting_row_found:
    #             if self.has_none_annotation(self.csv_annotation_values[new_file][2::]):
    #                 self.starting_row = dct_index
    #                 new_starting_row_found = True
    #     if not new_starting_row_found:
    #         self.starting_row = len(new_csv_annotations) - 1
    #     self.csv_annotation_values = new_csv_annotations

    def _next_image_clicked(self):
        """
        Move to the next image for annotating.

        If the last image is being annotated, write to csv. If the second
        image is being annotated, enable previous button.
        """

        self._annotator_model.set_previous_image_index(self._annotator_model.get_curr_img_index())
        # This dispatches a signal which updates  annotations dictionary and sets the next image
        self._annotator_model.set_curr_img_index(self._annotator_model.get_curr_img_index() + 1)
        self.annots.view.save_btn.setEnabled(True)

    def _prev_image_clicked(self):
        """
        Move to the previous image for annotating.

        If the first image is being annotated, disable button.
        """
        self._annotator_model.set_previous_image_index(self._annotator_model.get_curr_img_index())
        self._annotator_model.set_curr_img_index(self._annotator_model.get_curr_img_index() - 1)
        self.annots.view.save_btn.setEnabled(True)

    def _image_selected(self):
        """
        Record the annotations for the previously selected image and set current image.

        Called only when annotating un-blind and users select an image from the list.
        """
        self.annots.view.save_btn.setEnabled(True)
        self.annots.set_curr_img()

    def _exit_clicked(self):
        """Stop annotation if user confirms choice in popup."""
        proceed: bool = Popup.make_popup("Close this session? Your work will be saved.")
        if proceed:
            self._stop_annotating()

        self._images_view.update_num_files_label(self._annotator_model.get_num_images())

    def _save(self):
        """Save and write current annotations."""
        self.annots.save_annotations()
        self.annots.view.save_btn.setEnabled(False)

    def has_none_annotation(self, lst: List[Union[str, int, bool]]) -> bool:
        """
        Test if the given list has any empty string or None values.
        Returns
        -------
        bool
            True if null values are in the list.
        """

        if len(lst) < len(self._annotator_model.get_annotation_keys().keys()):
            return True
        else:
            for item in lst:
                if item is None or item == "":
                    return True
        return False



from pathlib import Path
import numpy as np
import dask.array as da
import napari_allencell_annotator
from napari_allencell_annotator.util.image_utils import ImageUtils


def test_get_dask_data_tiff() -> None:
    # ARRANGE
    test_path: Path = (
        Path(napari_allencell_annotator.__file__).parent / "_tests" / "assets" / "image_types" / "img.ome.tiff"
    )

    # ACT
    test_image: da.array = ImageUtils(test_path).get_image_dask_data()

    # ASSERT
    np.testing.assert_array_equal(test_image[0, 0, :, :, :, :], np.zeros((2, 2, 2, 2)))



from pathlib import Path

import numpy as np
import dask.array as da
from bioio import BioImage
import bioio_ome_tiff
import bioio_czi
import bioio_imageio
import bioio_ome_zarr


class ImageUtils:
    """
    Handles image display with BioImage

    Attributes
    ----------
    _image: BioImage
        An image to be displayed

    Methods
    -------
    get_dask_data(self) -> da.Array
        Returns the dask array of the image
    """

    def __init__(self, filepath: Path):

        extension: str = filepath.suffix

        self._image: BioImage
        if extension in [".tiff", ".tif"]:
            self._image = BioImage(filepath, reader=bioio_ome_tiff.Reader)
        elif extension == ".czi":
            self._image = BioImage(filepath, reader=bioio_czi.Reader)
        elif extension == ".zarr":
            self._image = BioImage(filepath, reader=bioio_ome_zarr.Reader)
        else:
            self._image = BioImage(filepath, reader=bioio_imageio.Reader)

    def get_image_dask_data(self) -> da.Array:
        """
        Returns image data as a dask array
        """
        return self._image.get_dask_stack()



from typing import Set

from PyQt5.QtCore import QSignalBlocker
from qtpy.QtWidgets import QListWidget
from qtpy.QtCore import Signal
from pathlib import Path

from napari_allencell_annotator.model.annotation_model import AnnotatorModel
from napari_allencell_annotator.widgets.file_item import FileItem


class FilesWidget(QListWidget):
    """
    A class used to create a QListWidget for files.

    Attributes
    ----------
    shuffled : bool
        a boolean, True if list is currently shuffled
    checked : Set[FileItem]
        a set of items that are currently checked

    Methods
    -------
    set_shuffled(shuffled : bool)
        Sets the list shuffled property.
    unhide_all()
        Displays the file names on all files in the list.
    clear_all()
        Clears all image data in the file widget.
    clear_for_shuff()
        Clears the list display.
    get_curr_row() -> int
        Returns current image row.
    add_item(file: Path, hidden: bool)
        Adds a file to the file widget.
    remove_item(item: ListItem)
        Removes the item from the file widget.
    """

    files_selected: Signal = Signal(bool)
    files_added: Signal = Signal(bool)

    def __init__(self, annotator_model: AnnotatorModel):
        QListWidget.__init__(self)
        self.checked: Set[FileItem] = set()
        # files_dict holds all image info file path -> [file name, FMS]
        # also holds the original insertion order in .keys()
        self.setCurrentItem(None)
        self._annotator_model = annotator_model

        self._annotator_model.image_changed.connect(self._handle_image_changed)
        self._annotator_model.images_shuffled.connect(self._handle_shuffle)
        self._annotator_model.image_set_added.connect(
            lambda: self._handle_shuffle(self._annotator_model.is_images_shuffled())
        )
        self._annotator_model.image_count_changed.connect(self._handle_image_count_change)
        self.currentItemChanged.connect(self._handle_file_item_changed)

    def _handle_image_changed(self):
        with QSignalBlocker(self):
            self.setCurrentItem(self.item(self._annotator_model.get_curr_img_index()))

            previous = self.item(self._annotator_model.get_previous_image_index())
            if previous is not None:
                previous.unhighlight()

            current = self.item(self._annotator_model.get_curr_img_index())
            if current is not None:
                current.highlight()

    def _handle_file_item_changed(self, curr_item: FileItem, prev_item: FileItem) -> None:
        """
        Update the model when the current item in the file widget is changed if it has not been updated.

        Parameters
        ----------
        curr_item:
            the current file item
        prev_item:
            the previous file item
        """
        self._annotator_model.set_previous_image_index(self.row(prev_item))
        self._annotator_model.set_curr_img_index(self.row(curr_item))

    def unhide_all(self) -> None:
        """Display the file names on all files in the list."""
        for i in range(self.count()):
            self.item(i).unhide()

    def get_curr_row(self) -> int:
        """
        Get the row of the currently selected image

        Returns
        -------
        int
            the current row.
        """
        if self.currentItem() is not None:
            return self.row(self.currentItem())
        else:
            return -1

    def _handle_shuffle(self, shuffled: bool) -> None:
        self._reset_list()
        if shuffled:
            # readd shuffled images to list
            for shuffled_img in self._annotator_model.get_shuffled_images():
                self.add_item(shuffled_img, hidden=True)  # add hidden when items shuffled.
        else:
            # readd unshuffled images to list
            for img in self._annotator_model.get_all_images():
                self.add_item(img)

    def _reset_list(self) -> None:
        """
        Reset the list of files
        """
        self.setCurrentItem(None)
        self.checked = set()
        self.clear()  # clear list

    def _handle_image_count_change(self, count: int) -> None:
        if count == 0:
            self._reset_list()

    def add_item(self, file: Path, hidden: bool = False) -> None:
        """
        Add a file to the file widget.

        Optional hidden parameter toggles file name visibility.

        Params
        -------
        file: Path
            a file path.
        hidden: bool
            file name visibility.
        """
        item: FileItem = FileItem(file, self, hidden)
        item.check.stateChanged.connect(lambda: self._check_evt(item))

    def remove_item(self, item: FileItem) -> None:
        """
        Remove the item from the file widget.

        Params
        -------
        item: FileItem
            an item to remove.
        """
        if item == self.currentItem():
            self.setCurrentItem(None)
        self.takeItem(self.row(item))

    def _check_evt(self, item: FileItem) -> None:
        """
        Update checked set and emit files_selected signal.

        Params
        -------
        item: FileItem
            the item that has been checked or unchecked.
        """
        if item.check.isChecked() and item not in self.checked:
            self.checked.add(item)
            if len(self.checked) == 1:
                self.files_selected.emit(True)
        elif not item.check.isChecked() and item in self.checked:
            self.checked.remove(item)
            if len(self.checked) == 0:
                self.files_selected.emit(False)

    def unhide_item_at(self, idx: int) -> None:
        self.unhide_item_at(idx)



from typing import Tuple, List, Optional, Any

from PyQt5.QtWidgets import QLayoutItem
from qtpy.QtWidgets import QLayout
from qtpy import QtWidgets
from qtpy.QtWidgets import (
    QListWidgetItem,
    QListWidget,
    QWidget,
    QHBoxLayout,
    QLineEdit,
    QComboBox,
    QLabel,
    QSpinBox,
    QGridLayout,
    QCheckBox,
)

from napari_allencell_annotator.model.combo_key import ComboKey
from napari_allencell_annotator.model.key import Key


class AnnotationItem(QListWidgetItem):
    """
    A class used to create custom annotation QListWidgetItems.
    """

    def __init__(self, parent: QListWidget):
        QListWidgetItem.__init__(self, parent)
        self.widget = QWidget()
        self.layout = QGridLayout()
        name_label = QLabel("Name:")

        self.name = QLineEdit()
        self.name.setPlaceholderText("Enter name")

        type_label = QLabel("Type:")
        self.type_selection_combo = QComboBox()
        self.type_selection_combo.addItems(["text", "number", "checkbox", "dropdown", "point"])
        self.name.setWhatsThis("name")
        self.type_selection_combo.setWhatsThis("type")
        self.name_widget = QWidget()
        self.name_layout = QHBoxLayout()
        self.check = QCheckBox()
        self.check.setToolTip("Check box to select this annotation for deletion.")

        self.name_layout.addWidget(self.check)
        self.name_layout.addWidget(name_label)
        self.name_layout.setSizeConstraint(QLayout.SetMinimumSize)
        self.name_widget.setLayout(self.name_layout)
        self.layout.addWidget(self.name_widget, 0, 0, 1, 1)
        self.layout.addWidget(self.name, 0, 1, 1, 2)
        self.layout.addWidget(type_label, 0, 3, 1, 1)
        self.layout.addWidget(self.type_selection_combo, 0, 4, 1, 2)
        self.default_label: QLabel = QLabel("Default:")
        self.default_text = QLineEdit()
        self.default_text.setPlaceholderText("Optional: Default Text")
        self.default_num = QSpinBox()
        self.default_num.setValue(2)
        self.default_check = QComboBox()
        self.default_check.addItems(["checked", "unchecked"])
        self.default_options_label = QLabel("Options:")
        self.default_options = QLineEdit()
        self.default_options.setPlaceholderText("Enter a comma separated list")

        self.default_options.setSizePolicy(QtWidgets.QSizePolicy.Policy.Ignored, QtWidgets.QSizePolicy.Policy.Preferred)
        self.default_options.setMinimumWidth(300)

        sp_retain = QtWidgets.QSizePolicy()
        sp_retain.setRetainSizeWhenHidden(True)
        self.default_options.setSizePolicy(sp_retain)
        self.default_options_label.setSizePolicy(sp_retain)

        self.layout.addWidget(self.default_label, 0, 6, 1, 1)
        self.layout.addWidget(self.default_text, 0, 7, 1, 2)
        self.layout.addWidget(self.default_options_label, 1, 1, 1, 1)
        self.layout.addWidget(self.default_options, 1, 2, 1, 7)
        self.default_options.hide()
        self.default_options_label.hide()

        self.layout.setContentsMargins(5, 5, 5, 15)

        self.widget.setLayout(self.layout)
        self.setSizeHint(self.widget.sizeHint())
        if parent is not None:
            parent.setItemWidget(self, self.widget)

        self.type_selection_combo.currentTextChanged.connect(self._type_changed)

    def fill_vals_text(self, name: str, default: str):
        """
        Fill in item name, default, and type for text.

        Parameters
        ----------
        name : str
            a name for the annotation
        default: str
            a default text value
        """
        self.type_selection_combo.setCurrentText("text")
        self.name.setText(name)
        self.default_text.setText(default)

    def fill_vals_number(self, name: str, default: int):
        """
        Fill in item name, default and type for number.

        Parameters
        ----------
        name : str
            a name for the annotation
        default: int
            a default number value
        """
        self.type_selection_combo.setCurrentText("number")
        self.name.setText(name)
        self.default_num.setValue(default)

    def fill_vals_check(self, name: str, default: bool):
        """
        Fill in name, default, and type for checkbox.

        Parameters
        ----------
        name : str
            a name for the annotation
        default: bool
            a bool default (True -> checked)
        """
        self.type_selection_combo.setCurrentText("checkbox")
        self.name.setText(name)
        if default:
            self.default_check.setCurrentText("checked")
        else:
            self.default_check.setCurrentText("unchecked")

    def fill_vals_list(self, name: str, default: str, options: List[str]):
        """
        Fill in item name, default, options, and type for dropdown.

        Parameters
        ----------
        name : str
            a name for the annotation
        default: str
            a default dropdown option
        options : List[str]
            a list of dropdown options
        """
        self.type_selection_combo.setCurrentText("dropdown")
        self.name.setText(name)
        self.default_text.setText(default)
        self.default_options.setText(", ".join(options))

    def fill_vals_point(self, name: str) -> None:
        """
        Fill in name for point.

        Parameters
        ----------
        name : str
            a name for the annotation
        """
        self.type_selection_combo.setCurrentText("point")
        self.name.setText(name)

    def _type_changed(self, text: str):
        """
        Render the widgets which correspond to the new type

        Parameters
        ----------
        text : str
            the new type selected.
        """
        default_item: QLayoutItem = self.layout.itemAtPosition(0, 7)
        if default_item is not None:
            default_widget: QWidget = default_item.widget()
            default_widget.setParent(None)
            self.layout.removeWidget(default_widget)

        if text == "text":
            self.default_label.show()
            self.default_options.hide()
            self.default_options_label.hide()
            self.layout.addWidget(self.default_text, 0, 7, 1, 2)

        elif text == "number":
            self.default_label.show()
            self.default_options.hide()
            self.default_options_label.hide()
            self.layout.addWidget(self.default_num, 0, 7, 1, 2)

        elif text == "checkbox":
            self.default_label.show()
            self.default_options.hide()
            self.default_options_label.hide()
            self.layout.addWidget(self.default_check, 0, 7, 1, 2)

        elif text == "dropdown":
            self.default_label.show()
            self.default_options.show()
            self.default_options_label.show()
            self.layout.addWidget(self.default_text, 0, 7, 1, 2)

        elif text == "point":
            self.default_label.hide()
            self.default_options.hide()
            self.default_options_label.hide()

    def get_data(self) -> Tuple[bool, str, Key, str]:
        """
        Highlight any invalid entries and return the data.

        Return True if all data is valid along with str name and a dictionary
        of annotation data.
        Return False if data is invalid, highlight the incorrect entries, and return
        an incomplete dictionary.

        Returns
        ------
        Tuple[bool, str, Dict]
            bool : True if entries are valid.
            str: name of annotation item
            Dict: annotation item data (type, default, options)
            str: error msg if applicable
        """

        # TODO: refactor this mess- bkim 7/1/20-24
        # bool valid if all annotation values are in the correct format
        error = ""
        valid: bool = True
        # test if annotation name is valid
        name: str = self.name.text()
        self._unhighlight(self.name)
        if name is None or name.isspace() or len(name) == 0:
            valid = False
            self.highlight(self.name)
            error = " Invalid Name. "

        key: Optional[Key] = None
        type: str = self.type_selection_combo.currentText()
        # dictionary of annotation type, to annotation keys
        default: Optional[Any] = None
        options: list[str] = []

        if type == "text" or type == "dropdown":
            # grab default text entry
            default = self.default_text.text()
            if default is None or len(default) == 0 or default.isspace():
                default = None
            else:
                default = default.strip()
                # default text exists
                default = default
            if type == "text":
                type = "string"
            else:
                type = "list"
                # type is options
                # comma separate list of options
                options_list = self.default_options.text()
                # unhighlight by default
                self._unhighlight(self.default_options)
                # if there is less than two options provided
                if options_list is None or len(options_list.split(",")) < 2:
                    valid = False
                    self.highlight(self.default_options)
                    error = error + " Must provide two dropdown options. "
                else:
                    options_list = [word.strip() for word in options_list.split(",")]
                    contained: bool = False
                    if default == "":
                        contained = True
                    for item in options_list:
                        # check each item in options
                        if len(item) == 0:
                            valid = False
                            self.highlight(self.default_options)
                            error = error + " Invalid options for dropdown. "
                            break
                        else:
                            if not contained and item == default:
                                contained = True
                    if default not in options_list:
                        error = f"Default value {default} is not in list of valid options"
                        valid = False
                    options = options_list
        elif type == "number":
            # number defaults are required by spinbox, always valid
            type = "number"
            default = self.default_num.value()
        elif type == "checkbox":
            # checkbox type default required by the drop down, always valid
            type = "bool"
            if self.default_check.currentText() == "checked":
                default = True
            else:
                default = False

        elif type == "point":
            type = "point"
            default = None

        if type == "list":
            key = ComboKey("list", options, default)
        else:
            key = Key(type, default)
        return valid, name, key, error

    def highlight(self, objct: QWidget):
        objct.setStyleSheet("""QLineEdit{border: 1px solid red}""")

    def _unhighlight(self, objct: QWidget):
        objct.setStyleSheet("""QLineEdit{}""")



from pathlib import Path
from typing import List

from napari_allencell_annotator.util.file_utils import FileUtils
import napari_allencell_annotator


def test_get_valid_images_sorted() -> None:
    # ARRANGE
    all_files: List[Path] = [
        Path("c.tiff"),
        Path("b.czi"),
        Path("a/raw.ome.zarr"),
        Path(".test.tiff"),
        Path("test.csv"),
        Path(napari_allencell_annotator.__file__).parent / "_tests" / "assets" / Path("valid_img_dir"),
        Path(napari_allencell_annotator.__file__).parent / "_tests" / "assets" / "zarr_dir",
    ]

    # ACT
    valid_files: List[Path] = FileUtils.get_valid_images_sorted(all_files)

    # ASSERT
    assert valid_files == [
        Path("a/raw.ome.zarr"),
        Path("b.czi"),
        Path("c.tiff"),
        Path(napari_allencell_annotator.__file__).parent / "_tests" / "assets" / "zarr_dir" / "raw.ome.zarr",
    ]


def test_get_file_name() -> None:
    # ACT
    tiff_file_name: str = FileUtils.get_file_name(Path("parent/test.tiff"))
    png_file_name: str = FileUtils.get_file_name(Path("parent/test.png"))
    jpeg_file_name: str = FileUtils.get_file_name(Path("parent/test.jpeg"))
    czi_file_name: str = FileUtils.get_file_name(Path("parent/test.czi"))
    zarr_file_name: str = FileUtils.get_file_name(Path("parent/test.ome.zarr"))

    # ASSERT
    assert tiff_file_name == "test.tiff"
    assert png_file_name == "test.png"
    assert jpeg_file_name == "test.jpeg"
    assert czi_file_name == "test.czi"
    assert zarr_file_name == "parent"



from PyQt5.QtWidgets import QListView, QAbstractItemView, QTreeView
from qtpy.QtWidgets import QFileDialog, QWidget
from napari_allencell_annotator._style import Style


class OmeZarrDirectoryOrFileDialog(QFileDialog):
    """
    A custom QFileDialog that allows the user to select files and zarr directories.
    """

    def __init__(self, parent: QWidget, title: str):
        super().__init__(parent, title)
        self.setStyleSheet(Style.get_stylesheet("main.qss"))
        self.currentChanged.connect(self._selected)
        self.setNameFilter("Directories and files (*)")
        self.setOption(QFileDialog.DontUseNativeDialog, True)

    def _selected(self, name: str) -> None:
        """
        Called whenever the user selects a new option in the File Dialog menu.
        """
        self.setFileMode(QFileDialog.Directory | QFileDialog.ExistingFiles)
        self.setNameFilter("Directories and files (*)")
        self.setOption(QFileDialog.DontUseNativeDialog, True)
        self.findChild(QListView, "listView").setSelectionMode(QAbstractItemView.ExtendedSelection)

        f_tree_view: QTreeView = self.findChild(QTreeView)
        if f_tree_view:
            f_tree_view.setSelectionMode(QAbstractItemView.ExtendedSelection)

    def accept(self) -> None:
        """
        Called whenever the user is done selecting files and directories.
        """
        self.setFileMode(QFileDialog.Directory | QFileDialog.ExistingFiles)
        self.setNameFilter("Directories and files (*)")
        self.setOption(QFileDialog.DontUseNativeDialog, True)
        self.findChild(QListView, "listView").setSelectionMode(QAbstractItemView.ExtendedSelection)

        f_tree_view: QTreeView = self.findChild(QTreeView)
        if f_tree_view:
            f_tree_view.setSelectionMode(QAbstractItemView.ExtendedSelection)

        super().done(QFileDialog.Accepted)



from pathlib import Path
from typing import Optional

from PyQt5.QtWidgets import QListWidgetItem

from napari_allencell_annotator.model.annotation_model import AnnotatorModel
from napari_allencell_annotator.view.i_viewer import IViewer
from napari_allencell_annotator.widgets.file_scrollable_popup import FileScrollablePopup
from napari_allencell_annotator.widgets.popup import Popup
from qtpy.QtWidgets import QFrame
from qtpy.QtCore import Qt

from qtpy.QtWidgets import (
    QLabel,
    QScrollArea,
    QGridLayout,
    QPushButton,
)


from napari_allencell_annotator.widgets.file_input import (
    FileInput,
    FileInputMode,
)
from napari_allencell_annotator.widgets.files_widget import FilesWidget, FileItem
from napari_allencell_annotator.util.file_utils import FileUtils
from napari_allencell_annotator._style import Style

from napari_allencell_annotator.util.image_utils import ImageUtils


class ImagesView(QFrame):
    """
    A class used to create a view for image file uploading and selecting.

    Attributes
    ----------
    viewer : IViewer
        a viewer where the plugin will be used
    ctrl : ImagesController
        a controller for the view

    Methods
    -------
    alert(alert:str)
        Displays the alert message on the napari viewer
    update_num_files_label(num_files:int)
        Updates num_files_label to show the current number of image files
    """

    def __init__(self, annotator_model: AnnotatorModel, viewer: IViewer):
        """
        Parameters
        ----------
        viewer : IViewer
            The viewer for the plugin
        """
        self._annotator_model: AnnotatorModel = annotator_model
        super().__init__()

        self.setStyleSheet(Style.get_stylesheet("main.qss"))

        self.label: QLabel = QLabel()
        self.label.setAlignment(Qt.AlignCenter)
        self.label.setText("Images")

        self.layout: QGridLayout = QGridLayout()
        self.layout.addWidget(self.label, 0, 0, 1, 4)

        self.input_dir: FileInput
        self.input_file: FileInput

        self.input_dir = FileInput(mode=FileInputMode.DIRECTORY, placeholder_text="Add a folder...")

        self.input_file = FileInput(mode=FileInputMode.FILE, placeholder_text="Add files...")
        self.layout.addWidget(self.input_dir, 1, 0, 1, 2)
        self.layout.addWidget(self.input_file, 1, 2, 1, 2)

        self.file_widget: FilesWidget = FilesWidget(self._annotator_model)

        self.scroll: QScrollArea = QScrollArea()
        self.scroll.setWidget(self.file_widget)
        self.scroll.setWidgetResizable(True)
        self.scroll.horizontalScrollBar().setEnabled(False)
        self.layout.addWidget(self.scroll, 2, 0, 10, 4)

        self.num_files_label: QLabel = QLabel("Image files:")
        self.layout.addWidget(self.num_files_label, 13, 0, 1, 4)

        self.shuffle: QPushButton = QPushButton("Shuffle and Hide")
        self.shuffle.setCheckable(True)

        self.shuffle.setEnabled(False)

        self.delete: QPushButton = QPushButton("Delete All")
        self.delete.setEnabled(False)

        self.layout.addWidget(self.shuffle, 14, 0, 1, 3)
        self.layout.addWidget(self.delete, 14, 3, 1, 1)

        self.setLayout(self.layout)

        self.viewer: IViewer = viewer
        self._connect_slots()

    def _connect_slots(self) -> None:
        """Connect signals to slots."""
        self.input_dir.dir_selected.connect(self._add_sorted_valid_images_in_dir)
        self.input_file.files_selected.connect(self._add_sorted_valid_images_in_files)
        self.shuffle.clicked.connect(self._handle_shuffle_clicked)
        self.delete.clicked.connect(self._handle_delete_clicked)
        self.file_widget.files_selected.connect(self._toggle_delete_button_text)

        self._annotator_model.image_changed.connect(self._display_img)
        self._annotator_model.image_count_changed.connect(self._handle_image_count_changed)
        self._annotator_model.images_shuffled.connect(self._handle_shuffle_ui)
        self._annotator_model.images_shuffled.connect(self.viewer.clear_layers)

    def _handle_shuffle_ui(self, checked: bool) -> None:
        """
        Update shuffle button text to reflect toggle state.

        Parameters
        ----------
        checked : bool
            Toggle state of shuffle button.
        """
        if checked:
            self.shuffle.setText("Unhide")
            self.disable_add_buttons()
        else:
            self.shuffle.setText("Shuffle and Hide")
            self.enable_add_buttons()

    def reset_buttons(self) -> None:
        """
        Reset buttons to pre-annotation settings

        Disable delete, add, and shuffle buttons.
        """
        self._toggle_delete_button_text(False)
        self.shuffle.setChecked(False)
        self._disable_delete_button()
        self._disable_shuffle_button()
        self.enable_add_buttons()

    def enable_add_buttons(self) -> None:
        """Enables add file and add directory buttons."""
        self.input_dir.toggle(True)
        self.input_file.toggle(True)

    def disable_add_buttons(self) -> None:
        """Disables add file and add directory buttons."""
        self.input_dir.toggle(False)
        self.input_file.toggle(False)

    def _toggle_delete_button_text(self, checked: bool) -> None:
        """
        Enable delete button when files are checked.

        Parameters
        ----------
        checked : bool
        """
        if checked:
            self.delete.setText("Delete Selected")
        elif not checked:
            self.delete.setText("Delete All")

    def _enable_delete_button(self) -> None:
        """Enable delete button"""
        self.delete.setToolTip("Check box on the right \n to select files for deletion")
        self.delete.setText("Delete All")
        self.delete.setEnabled(True)

    def _disable_delete_button(self) -> None:
        """Disable delete button"""
        self.delete.setToolTip(None)
        self.delete.setEnabled(False)

    def _enable_shuffle_button(self) -> None:
        """Enable shuffle button"""
        self.shuffle.setEnabled(True)

    def _disable_shuffle_button(self) -> None:
        """Disable shuffle button"""
        self.shuffle.setEnabled(False)

    def _display_img(self) -> None:
        """
        Display the current image in napari.

        Parameters
        ----------
        current: FileItem
            Current file
        previous: FileItem
            Previous file
        """
        self.viewer.clear_layers()

        current: Path = self._annotator_model.get_curr_img()
        if current is not None:
            self.viewer.add_image(ImageUtils(current).get_image_dask_data())

    def update_num_files_label(self, num_files: int) -> None:
        """
        Update num_files_label to show the number of image files

        Parameters
        ----------
        num_files: int
            The number of image files
        """
        self.num_files_label.setText(f"Image files: {num_files}")

    def _add_sorted_valid_images_in_files(self, file_list: list[Path]) -> None:
        all_sorted_valid_files: list[Path] = FileUtils.get_valid_images_sorted(file_list)
        self._add_selected_files(all_sorted_valid_files)

        if len(all_sorted_valid_files) != len([file for file in file_list if not file.name.startswith(".")]):
            self.viewer.alert("Unsupported file type(s)")

    def _add_sorted_valid_images_in_dir(self, dir_path: Path) -> None:
        """
        Adds all files in a directory to the GUI.

        Parameters
        ----------
        dir_path : Path
            The directory path
        """
        # get both files and folders
        dir_files = list(dir_path.glob("*"))
        all_sorted_valid_files_in_dir: list[Path] = FileUtils.get_valid_images_sorted(dir_files)

        if len(all_sorted_valid_files_in_dir) < 1:
            self.viewer.alert("Folder is empty")
        else:
            self._add_selected_files(all_sorted_valid_files_in_dir)

        if len(all_sorted_valid_files_in_dir) != len([file for file in dir_files if not file.name.startswith(".")]):
            self.viewer.alert("Unsupported file type(s)")

    def add_new_item(self, file: Path, hidden: Optional[bool] = False) -> None:
        """
        Add a new image to the model and the file widget.

        Optional hidden parameter toggles file name visibility. This function emits a files_added signal when this is
        the first file added and updates num_files_label.

        Parameters
        ----------
        file: Path
            The file path of a new image to be added
        hidden : Optional[bool]
            File name visibility
        """
        # TODO update model, which dispatches event, and have filewidget react to it.
        self.file_widget.add_item(file, hidden)
        self._annotator_model.add_image(file)  # update model

    def _add_selected_files(self, file_list: list[Path]) -> None:
        """
        Adds all selected files to the GUI and update state

        Parameters
        ----------
        file_list : List[Path]
            The list of files
        """
        for file_path in file_list:
            if file_path not in self._annotator_model.get_all_images():
                self.add_new_item(file_path)

    def _handle_shuffle_clicked(self) -> None:
        """
        Shuffle file order and hide file names if checked.
        Return files to original order and names if unchecked.

        Side effect: set file_widget.shuffled_files_dict to a new order dict or {} if list is unshuffled.

        Parameters
        ----------
        checked : bool
            Toggle state of the shuffle button.
        """
        new_toggle_state: bool = not self._annotator_model.is_images_shuffled()

        # self._update_shuff_text(new_toggle_state)
        if new_toggle_state:
            # Switching to shuffle: on
            # self.disable_add_buttons()
            # Set shuffled_files field in model, will emit event to have UI react to shuffled files.
            self._annotator_model.set_shuffled_images(
                FileUtils.shuffle_file_list(self._annotator_model.get_all_images())
            )
        else:
            # Switching to shuffle: off
            # self.enable_add_buttons()  # re-enable image adding
            # Set shuffled_files to None, indicating we've unshuffled the images
            self._annotator_model.set_shuffled_images(None)

    def _handle_delete_clicked(self) -> None:
        """
        Ask user to approve a list of files to delete and remove image files from the model and the file widget.

        If at least one file is checked, delete only selected files. Otherwise, delete all files.
        """
        # TODO can we do this with signals instead?
        if len(self.file_widget.checked) > 0:
            proceed: bool = FileScrollablePopup.make_popup(
                "Delete these files from the list?", self.file_widget.checked
            )
            if proceed:
                self.delete_checked()
        else:
            proceed: bool = Popup.make_popup("Remove all images?")
            if proceed:
                self.clear_all()

    def delete_checked(self) -> None:
        """
        Delete the checked items from the model and the file widget.
        """
        for item in self.file_widget.checked:
            self.remove_image(item)

        self.file_widget.checked.clear()

    def remove_image(self, item: FileItem) -> None:
        """
        Remove an image file from the model and the file widget.

        This function emits a files_added signal when the item to remove is the only item and updates num_files_label.

        Parameters
        ----------
        item: FileItem
            An item to be removed.
        """
        # TODO when we delete from the model, connect file widget so that it deletes that entry itself without
        # us explicitly calling remove_item on it
        if item.file_path in self._annotator_model.get_all_images():
            if self.file_widget.currentItem() == item:
                self.viewer.clear_layers()

            self.file_widget.remove_item(item)
            self._annotator_model.remove_image(item.file_path)
            self.update_num_files_label(self._annotator_model.get_num_images())

    def clear_all(self) -> None:
        """
        Clear all image data from the model and the file widget.
        """
        self._annotator_model.clear_all_images()  # clear model
        # self._annotator_model.set_shuffled_images(None)  # clear shuffled images, if any

    def start_annotating(self) -> None:
        """Set current item to the one at row."""
        count: int = self.file_widget.count()
        for x in range(count):
            file_item: Optional[QListWidgetItem] = self.file_widget.item(x)
            file_item.hide_check()
        else:
            self.viewer.alert("No files to annotate")

    def hide_image_paths(self) -> None:
        """
        Hide image paths (if shuffle is selected for annotations)
        """
        self.input_dir.hide()
        self.input_file.hide()
        self.shuffle.hide()
        self.delete.hide()

    def _handle_image_count_changed(self, count) -> None:
        self.update_num_files_label(count)
        if count > 0:
            self._enable_delete_button()
            self._enable_shuffle_button()
        else:
            self.viewer.clear_layers()
            self.reset_buttons()

    def stop_annotating(self):
        """Clear file widget and reset buttons."""
        self.file_widget.clear_all()
        self.view.reset_buttons()
